#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import json
import secrets
import unittest
from os import path

from btclib import bip39


class TestBIP39(unittest.TestCase):
    def test_bip39(self):
        lang = "en"
        raw_entr = bytes.fromhex("0000003974d093eda670121023cd0000")
        mnemonic = bip39.mnemonic_from_entropy(raw_entr, lang)
        self.assertEqual(
            mnemonic,
            "abandon abandon atom trust ankle walnut "
            "oil across awake bunker divorce abstract",
        )
        r = bip39.entropy_from_mnemonic(mnemonic, lang)
        size = (len(r) + 7) // 8
        r = int(r, 2).to_bytes(size, byteorder="big")
        self.assertEqual(r, raw_entr)

        # mnemonic with wrong number of words
        wrong_mnemonic = mnemonic + " abandon"
        self.assertRaises(ValueError, bip39.entropy_from_mnemonic, wrong_mnemonic, lang)
        # bip39_entropy_from_mnemonic(wrong_mnemonic, lang)

        # invalid mnemonic checksum
        wr_m = (
            "abandon abandon atom  trust  ankle   walnut "
            "oil     across  awake bunker divorce walnut"
        )
        self.assertRaises(ValueError, bip39.entropy_from_mnemonic, wr_m, lang)
        # bip39_entropy_from_mnemonic(wrong_mnemonic, lang)

        # Invalid number of bits (130) for BIP39 entropy; must be in ...
        binstr_entropy = "01" * 65  # 130 bits
        self.assertRaises(ValueError, bip39._entropy_checksum, binstr_entropy)
        # bip39._entropy_checksum(binstr_entropy)

    def test_vectors(self):
        """BIP39 test vectors
           https://github.com/trezor/python-mnemonic/blob/master/vectors.json
        """
        fname = "bip39_test_vectors.json"
        filename = path.join(path.dirname(__file__), "test_data", fname)
        with open(filename, "r") as f:
            test_vectors = json.load(f)["english"]
        for test_vector in test_vectors:
            lang = "en"
            entropy = bytes.fromhex(test_vector[0])
            mnemonic = bip39.mnemonic_from_entropy(entropy, lang)
            self.assertEqual(mnemonic.split(), test_vector[1].split())

            raw_entr = bip39.entropy_from_mnemonic(mnemonic, lang)
            size = (len(raw_entr) + 7) // 8
            raw_entr = int(raw_entr, 2).to_bytes(size, byteorder="big")
            self.assertEqual(raw_entr, entropy)

            seed = bip39.seed_from_mnemonic(mnemonic, "TREZOR").hex()
            self.assertEqual(seed, test_vector[2])

            # test_vector[3], i.e. the bip32 master private key from seed,
            # has been tested in bip32, as it does not belong here

    def test_zeroleadingbit(self):
        # it should not throw an error
        bip39.mnemonic_from_entropy(secrets.randbits(127), "en")


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Number theory and modular arithmetic functions.

Implementations originally from
https://en.wikibooks.org/wiki/Algorithm_Implementation/Mathematics/Extended_Euclidean_algorithm
and
https://codereview.stackexchange.com/questions/43210/tonelli-shanks-algorithm-implementation-of-prime-modular-square-root/43267
with the following modifications:

* type annotated python3
* minor improvements
* added extensive unit test
"""

from typing import Tuple


def xgcd(a: int, b: int) -> Tuple[int, int, int]:
    """Return (g, x, y) such that a*x + b*y = g = gcd(x, y).

    based on Extended Euclidean Algorithm, see
    https://en.wikibooks.org/wiki/Algorithm_Implementation/Mathematics/Extended_Euclidean_algorithm
    """

    x0, x1, y0, y1 = 0, 1, 1, 0
    while a != 0:
        q, b, a = b // a, a, b % a
        y0, y1 = y1, y0 - q * y1
        x0, x1 = x1, x0 - q * x1
    return b, x0, y0


def mod_inv(a: int, m: int) -> int:
    """Return the inverse of a (mod m). m does not have to be a prime.

    Based on Extended Euclidean Algorithm, see:
    https://en.wikibooks.org/wiki/Algorithm_Implementation/Mathematics/Extended_Euclidean_algorithm
    """

    a %= m
    g, x, _ = xgcd(a, m)
    if g == 1:
        return x % m
    raise ValueError(f"{hex(a)} has no inverse (mod {hex(m)})")


def legendre_symbol(a, p) -> int:
    """Compute the Legendre symbol a|p using Euler's criterion.

    p is a prime, a is relatively prime to p (if p divides a,
    then a|p = 0).
    It returns 1 if a has a square root modulo p, -1 otherwise.

    https://codereview.stackexchange.com/questions/43210/tonelli-shanks-algorithm-implementation-of-prime-modular-square-root/43267
    """

    ls = pow(a, p >> 1, p)
    return -1 if ls == p - 1 else ls


def mod_sqrt(a: int, p: int) -> int:
    """Return a quadratic residue (mod p) of a; p must be a prime.

    Solve the equation:
        x^2 = a mod p

    and return x. Note that p - x is also a root.

    If a simple solution is not available for p,
    then the Tonelli-Shanks algorithm is used.

    https://codereview.stackexchange.com/questions/43210/tonelli-shanks-algorithm-implementation-of-prime-modular-square-root/43267
    """

    a %= p

    # Simple cases
    if p % 4 == 3:  # secp256k1 case
        # inverse candidate is pow(a, (p + 1) // 4, p)
        x = pow(a, (p >> 2) + 1, p)
        if x * x % p == a:
            return x
        raise ValueError(f"{hex(a)} has no root (mod {hex(p)})")
    elif p % 8 == 5:
        # inverse candidate is pow(a, (p + 3) // 8, p)
        x = pow(a, (p >> 3) + 1, p)
        if x * x % p == a:
            return x
        else:
            # inverse candidate
            x = x * pow(2, p >> 2, p) % p
            if x * x % p == a:
                return x
        raise ValueError(f"{hex(a)} has no root (mod {hex(p)})")
    elif a == 0 or p == 2:
        return a

    # Check solution existence for odd primes
    if legendre_symbol(a, p) != 1:
        raise ValueError(f"{hex(a)} has no root (mod {hex(p)})")

    # Factor p-1 on the form q * 2^s (with Q odd)
    q, s = p - 1, 0
    while q & 1 == 0:
        s += 1
        q >>= 1

    # Select a z which is a quadratic non resudue modulo p
    z = 1
    while legendre_symbol(z, p) != -1:
        z += 1
    c = pow(z, q, p)

    # Search for a solution
    x = pow(a, (q + 1) // 2, p)
    t = pow(a, q, p)
    m = s
    while t != 1:
        # Find the lowest i such that t^(2^i) = 1
        t2i = t
        for i in range(1, m):
            t2i = t2i * t2i % p
            if t2i == 1:
                break

        # Update next value to iterate
        b = pow(c, 1 << (m - i - 1), p)
        x = (x * b) % p
        c = (b * b) % p
        t = (t * c) % p
        m = i

    return x



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Assorted conversion utilities.

Most conversions from SEC 1 v.2 2.3 are included.

https://www.secg.org/sec1-v2.pdf
"""

import hashlib
from typing import Iterable, Optional, Union

from .alias import Integer, Octets

# hexstr_from_bytes is not needed!!
# def hexstr_from_bytes(byte_str: bytes) -> str:
#    return byte_str.hex()


NoneOneOrMoreInt = Optional[Union[int, Iterable[int]]]


def bytes_from_octets(o: Octets, out_size: NoneOneOrMoreInt = None) -> bytes:
    """Return bytes from a hex-string, stripping leading/trailing spaces.

    If the input is not a string, then it goes untouched.
    Optionally, it also ensures required output size.
    """

    if isinstance(o, str):  # hex string
        o = bytes.fromhex(o)

    if (
        out_size is None
        or isinstance(out_size, int)
        and len(o) == out_size
        or isinstance(out_size, Iterable)
        and len(o) in out_size
    ):
        return o

    m = f"Invalid size: {len(o)} bytes instead of {out_size}"
    raise ValueError(m)


def int_from_integer(i: Integer) -> int:
    if isinstance(i, int):
        return i
    elif isinstance(i, bytes):
        return int.from_bytes(i, "big")
    return int(i, 16)


def int_from_bits(o: Octets, nlen: int) -> int:
    """Return the leftmost nlen bits.

    Take as input a sequence of blen bits and calculate a
    non-negative integer i that is less than 2^nlen according to
    SEC 1 v.2 section 4.1.3 (5).
    Note that an additional reduction modulo n would be required
    to ensure that 0 < i < n.

    int_from_bits is not the reverse of i.to_bytes, even
    for input sequences of length nlen: i.to_bytes will add some
    bits on the left, while int_from_bits will discard some bits on the
    right. i.to_bytes is the reverse of int_from_bits only when
    nlen is a multiple of 8 and bit sequences already have length nlen.
    See https://tools.ietf.org/html/rfc6979#section-2.3.5.
    """

    o = bytes_from_octets(o)
    i = int.from_bytes(o, byteorder="big")

    blen = len(o) * 8  # bits
    n = (blen - nlen) if blen >= nlen else 0
    return i >> n


def sha256(o: Octets) -> bytes:
    """Return SHA256(*) of the input octet sequence."""

    o = bytes_from_octets(o)
    return hashlib.sha256(o).digest()


def hash160(o: Octets) -> bytes:
    """Return RIPEMD160(SHA256(*)) of the input octet sequence."""

    t = sha256(o)
    return hashlib.new("ripemd160", t).digest()


def hash256(o: Octets) -> bytes:
    """Return SHA256(SHA256(*)) of the input octet sequence."""

    t = sha256(o)
    return hashlib.sha256(t).digest()


def ensure_is_power_of_two(n: int, var_name: str = None) -> None:
    # http://www.graphics.stanford.edu/~seander/bithacks.html
    if n & (n - 1) != 0:
        raise ValueError(f"{var_name} ({n}) must be a power of two")



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Bitcoin Script.

https://en.bitcoin.it/wiki/Script

Scripts are represented by List[Token], where Token = Union[int, str, bytes]:

* int [-1, 16] are shorcuts for 'OP_1NEGATE', 'OP_0' - 'OP_16'
* str are for opcodes (e.g. 'OP_HASH160') or hexstring data
* bytes are for data (but integers are often casted to int)
"""

from io import BytesIO
from typing import BinaryIO, List, Union

from . import varint
from .alias import Octets, Token
from .utils import bytes_from_octets

SIGHASH_ALL = 1
SIGHASH_NONE = 2
SIGHASH_SINGLE = 3
SIGHASH_ANYONECANPAY = 0x80

SIGHASHES = [
    bytes([SIGHASH_ALL]),
    bytes([SIGHASH_NONE]),
    bytes([SIGHASH_SINGLE]),
    bytes([SIGHASH_ANYONECANPAY + SIGHASH_ALL]),
    bytes([SIGHASH_ANYONECANPAY + SIGHASH_NONE]),
    bytes([SIGHASH_ANYONECANPAY + SIGHASH_SINGLE]),
]

OP_CODES = {
    # Constants
    "OP_0": b"\x00",
    "OP_FALSE": b"\x00",
    "OP_PUSHDATA1": b"\x4c",
    "OP_PUSHDATA2": b"\x4d",
    "OP_PUSHDATA4": b"\x4e",
    "OP_1NEGATE": b"\x4f",
    "OP_1": b"\x51",
    "OP_TRUE": b"\x51",
    "OP_2": b"\x52",
    "OP_3": b"\x53",
    "OP_4": b"\x54",
    "OP_5": b"\x55",
    "OP_6": b"\x56",
    "OP_7": b"\x57",
    "OP_8": b"\x58",
    "OP_9": b"\x59",
    "OP_10": b"\x5a",
    "OP_11": b"\x5b",
    "OP_12": b"\x5c",
    "OP_13": b"\x5d",
    "OP_14": b"\x5e",
    "OP_15": b"\x5f",
    "OP_16": b"\x60",
    # Flow control
    "OP_NOP": b"\x61",
    "OP_IF": b"\x63",
    "OP_NOTIF": b"\x64",
    "OP_ELSE": b"\x67",
    "OP_ENDIF": b"\x68",
    "OP_VERIFY": b"\x69",
    "OP_RETURN": b"\x6a",
    # Stack
    "OP_TOALTSTACK": b"\x6b",
    "OP_FROMALTSTACK": b"\x6c",
    "OP_IFDUP": b"\x73",
    "OP_DEPTH": b"\x74",
    "OP_DROP": b"\x75",
    "OP_DUP": b"\x76",
    "OP_NIP": b"\x77",
    "OP_OVER": b"\x78",
    "OP_PICK": b"\x79",
    "OP_ROLL": b"\x7a",
    "OP_ROT": b"\x7b",
    "OP_SWAP": b"\x7c",
    "OP_TUCK": b"\x7d",
    "OP_2DROP": b"\x6d",
    "OP_2DUP": b"\x6e",
    "OP_3DUP": b"\x6f",
    "OP_2OVER": b"\x70",
    "OP_2ROT": b"\x71",
    "OP_2SWAP": b"\x72",
    # Splice
    "OP_SIZE": b"\x82",
    # Bitwise logic
    "OP_EQUAL": b"\x87",
    "OP_EQUALVERIFY": b"\x88",
    # Arithmetic
    "OP_1ADD": b"\x8b",  # without OP_, 1ADD would be a number
    "OP_1SUB": b"\x8c",
    "OP_NEGATE": b"\x8f",
    "OP_ABS": b"\x90",
    "OP_NOT": b"\x91",
    "OP_0NOTEQUAL": b"\x92",
    "OP_ADD": b"\x93",
    "OP_SUB": b"\x94",
    "OP_BOOLAND": b"\x9a",
    "OP_BOOLOR": b"\x9b",
    "OP_NUMEQUAL": b"\x9c",
    "OP_NUMEQUALVERIFY": b"\x9d",
    "OP_NUMNOTEQUAL": b"\x9e",
    "OP_LESSTHAN": b"\x9f",
    "OP_GREATERTHAN": b"\xa0",
    "OP_LESSTHANOREQUAL": b"\xa1",
    "OP_GREATERTHANOREQUAL": b"\xa2",
    "OP_MIN": b"\xa3",
    "OP_MAX": b"\xa4",
    "OP_WITHIN": b"\xa5",
    # Crypto
    "OP_RIPEMD160": b"\xa6",
    "OP_SHA1": b"\xa7",
    "OP_SHA256": b"\xa8",
    "OP_HASH160": b"\xa9",
    "OP_HASH256": b"\xaa",
    "OP_CODESEPARATOR": b"\xab",
    "OP_CHECKSIG": b"\xac",
    "OP_CHECKSIGVERIFY": b"\xad",
    "OP_CHECKMULTISIG": b"\xae",
    "OP_CHECKMULTISIGVERIFY": b"\xaf",
    # Locktime
    "OP_NOP2": b"\xb1",
    "OP_CHECKLOCKTIMEVERIFY": b"\xb1",
    "OP_NOP3": b"\xb2",
    "OP_CHECKSEQUENCEVERIFY": b"\xb2",
    # Reserved words
    "OP_NOP1": b"\xb0",
    "OP_NOP4": b"\xb3",
    "OP_NOP5": b"\xb4",
    "OP_NOP6": b"\xb5",
    "OP_NOP7": b"\xb6",
    "OP_NOP8": b"\xb7",
    "OP_NOP9": b"\xb8",
    "OP_NOP10": b"\xb9",
}

OP_CODE_NAMES = {
    0: "OP_0",
    76: "OP_PUSHDATA1",
    77: "OP_PUSHDATA2",
    78: "OP_PUSHDATA4",
    79: "OP_1NEGATE",
    81: "OP_1",
    82: "OP_2",
    83: "OP_3",
    84: "OP_4",
    85: "OP_5",
    86: "OP_6",
    87: "OP_7",
    88: "OP_8",
    89: "OP_9",
    90: "OP_10",
    91: "OP_11",
    92: "OP_12",
    93: "OP_13",
    94: "OP_14",
    95: "OP_15",
    96: "OP_16",
    97: "OP_NOP",
    99: "OP_IF",
    100: "OP_NOTIF",
    103: "OP_ELSE",
    104: "OP_ENDIF",
    105: "OP_VERIFY",
    106: "OP_RETURN",
    107: "OP_TOALTSTACK",
    108: "OP_FROMALTSTACK",
    109: "OP_2DROP",
    110: "OP_2DUP",
    111: "OP_3DUP",
    112: "OP_2OVER",
    113: "OP_2ROT",
    114: "OP_2SWAP",
    115: "OP_IFDUP",
    116: "OP_DEPTH",
    117: "OP_DROP",
    118: "OP_DUP",
    119: "OP_NIP",
    120: "OP_OVER",
    121: "OP_PICK",
    122: "OP_ROLL",
    123: "OP_ROT",
    124: "OP_SWAP",
    125: "OP_TUCK",
    130: "OP_SIZE",
    135: "OP_EQUAL",
    136: "OP_EQUALVERIFY",
    139: "OP_1ADD",
    140: "OP_1SUB",
    143: "OP_NEGATE",
    144: "OP_ABS",
    145: "OP_NOT",
    146: "OP_0NOTEQUAL",
    147: "OP_ADD",
    148: "OP_SUB",
    154: "OP_BOOLAND",
    155: "OP_BOOLOR",
    156: "OP_NUMEQUAL",
    157: "OP_NUMEQUALVERIFY",
    158: "OP_NUMNOTEQUAL",
    159: "OP_LESSTHAN",
    160: "OP_GREATERTHAN",
    161: "OP_LESSTHANOREQUAL",
    162: "OP_GREATERTHANOREQUAL",
    163: "OP_MIN",
    164: "OP_MAX",
    165: "OP_WITHIN",
    166: "OP_RIPEMD160",
    167: "OP_SHA1",
    168: "OP_SHA256",
    169: "OP_HASH160",
    170: "OP_HASH256",
    171: "OP_CODESEPARATOR",
    172: "OP_CHECKSIG",
    173: "OP_CHECKSIGVERIFY",
    174: "OP_CHECKMULTISIG",
    175: "OP_CHECKMULTISIGVERIFY",
    176: "OP_NOP1",
    177: "OP_CHECKLOCKTIMEVERIFY",
    178: "OP_CHECKSEQUENCEVERIFY",
    179: "OP_NOP4",
    180: "OP_NOP5",
    181: "OP_NOP6",
    182: "OP_NOP7",
    183: "OP_NOP8",
    184: "OP_NOP9",
    185: "OP_NOP10",
}


def _op_pushdata(data: Octets) -> bytes:
    """Convert to canonical push: OP_PUSHDATA (if needed) | length | data.

    According to standardness rules (BIP-62) the
    minimum possible PUSHDATA operator must be used.
    Byte vectors on the stack are not allowed to be more than 520 bytes long.
    """

    data = bytes_from_octets(data)
    r = b""
    length = len(data)
    if length < 75:  # 1-byte-length
        r += length.to_bytes(1, byteorder="little")
    elif length < 256:  # OP_PUSHDATA1 | 1-byte-length
        r += OP_CODES["OP_PUSHDATA1"]
        r += length.to_bytes(1, byteorder="little")
    elif length < 521:  # OP_PUSHDATA2 | 2-byte-length
        r += OP_CODES["OP_PUSHDATA2"]
        r += length.to_bytes(2, byteorder="little")
    else:
        # because of the 520 bytes limit
        # there is no need to use OP_PUSHDATA4
        # r += OP_CODES['OP_PUSHDATA4']
        # r += length.to_bytes(4, byteorder='little')
        raise ValueError(f"Cannot push {length} bytes on the stack")
    r += data
    return r


def _op_int(token: int) -> bytes:
    # Short 1-byte opcodes exist
    # to push numbers in [-1, 16]
    if token >= 0 and token <= 16:
        return OP_CODES["OP_" + str(token)]
    elif token == -1:
        return OP_CODES["OP_1NEGATE"]
    # Pushing any other number requires an
    # explicit push operation of its bytes encoding
    # FIXME: negative numbers?
    else:
        nbytes = (token.bit_length() + 7) // 8
        data = token.to_bytes(nbytes, byteorder="little")
        return _op_pushdata(data)


def _op_str(token: str) -> bytes:
    token = token.strip()
    token = token.upper()
    if token in OP_CODES:
        return OP_CODES[token]
    try:
        data = bytes.fromhex(token)
    except Exception:
        raise ValueError(f"Invalid string token: {token}")
    return _op_pushdata(data)


def encode(script: List[Token]) -> bytes:
    r = b""
    for token in script:
        if isinstance(token, int):
            r += _op_int(token)
        elif isinstance(token, str):
            r += _op_str(token)
        elif isinstance(token, bytes):
            r += _op_pushdata(token)
        else:
            raise ValueError(f"Unmanaged {type(token)} token type")
    return r


def decode(script: Octets) -> List[Token]:

    script = bytes_from_octets(script)

    # initialize the result list
    r: List[Token] = []

    length = len(script)
    s = BytesIO(script)  # TODO: avoid unnecessary streamification
    counter = 0
    while counter < length:
        # get one byte and convert it to an integer
        i = s.read(1)[0]
        counter += 1
        if i == 0:
            # numeric value 0 (OP_0)
            # r.append(OP_CODE_NAMES[i])
            r.append(i)
        elif i == 79:
            # numeric value -1 (OP_1NEGATE)
            r.append(-1)
        elif i > 80 and i < 97:
            # numeric values 1-16 (OP_1-OP_16)
            # r.append(OP_CODE_NAMES[i])
            r.append(i - 80)
        elif i < 76:
            # 1-byte-data-length | data
            data = s.read(i)
            counter += i
            r.append(data.hex().upper())
        elif i == 76:
            # OP_PUSHDATA1 | 1-byte-data-length | data
            data_length = int.from_bytes(s.read(1), byteorder="little")
            data = s.read(data_length)
            counter += 1 + data_length
            r.append(data.hex().upper())
        elif i == 77:
            # OP_PUSHDATA2 | 2-byte-data-length | data
            data_length = int.from_bytes(s.read(2), byteorder="little")
            data = s.read(data_length)
            counter += 2 + data_length
            r.append(data.hex().upper())
        elif i == 78:
            # OP_PUSHDATA4 | 4-byte-data-length | data
            data_length = int.from_bytes(s.read(4), byteorder="little")
            data = s.read(data_length)
            counter += 4 + data_length
            r.append(data.hex().upper())
        else:
            # OP_CODE
            r.append(OP_CODE_NAMES[i])

    return r


def serialize(script: List[Token]) -> bytes:
    r = encode(script)
    # prepend length as varint
    length = len(r)
    return varint.encode(length) + r


def deserialize(stream: Union[BinaryIO, Octets]) -> List[Token]:

    if isinstance(stream, str):
        stream = bytes_from_octets(stream)

    if isinstance(stream, bytes):
        stream = BytesIO(stream)

    length = varint.decode(stream)
    script = stream.read(length)
    return decode(script)



#!/usr/bin/env python3

# Copyright (C) 2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

from typing import Optional, Tuple

from . import bip32
from .alias import BIP32Key, PrvKey, String
from .base58 import b58decode
from .curve import Curve
from .curves import secp256k1
from .network import (
    NETWORKS,
    network_from_key_value,
    network_from_xkeyversion,
    xprvversions_from_network,
)
from .utils import bytes_from_octets


def int_from_prvkey(prvkey: PrvKey, ec: Curve = secp256k1) -> int:
    """Return a verified-as-valid private key integer.

    It supports:

    - WIF (bytes or string)
    - BIP32 extended keys (bytes, string, or BIP32KeyDict)
    - SEC Octets (bytes or hex-string, with 02, 03, or 04 prefix)
    - integer (native int or hex-strin)

    Network and compressed informations from the input key
    are not used.
    """

    if isinstance(prvkey, int):
        q = prvkey
    elif isinstance(prvkey, dict):
        q, network, _ = _prvkeyinfo_from_xprvwif(prvkey)
        # q has been validated on the xprv/wif network
        ec2 = NETWORKS[network]["curve"]
        assert ec == ec2, f"ec / network ({network}) mismatch"
        return q
    else:
        try:
            q, network, _ = _prvkeyinfo_from_xprvwif(prvkey)
        except Exception:
            pass
        else:
            # q has been validated on the xprv/wif network
            ec2 = NETWORKS[network]["curve"]
            assert ec == ec2, f"ec / network ({network}) mismatch"
            return q

        prvkey = bytes_from_octets(prvkey, ec.nsize)
        q = int.from_bytes(prvkey, "big")

    if not 0 < q < ec.n:
        raise ValueError(f"Private key {hex(q).upper()} not in [1, n-1]")

    return q


PrvKeyInfo = Tuple[int, str, bool]


def _prvkeyinfo_from_wif(
    wif: String, network: Optional[str] = None, compressed: Optional[bool] = None
) -> PrvKeyInfo:
    """Return private key tuple(int, compressed, network) from a WIF.

    WIF is always compressed and includes network information:
    here the 'network, compressed' input parameters are passed
    only to allow consistency checks.
    """

    if isinstance(wif, str):
        wif = wif.strip()

    payload = b58decode(wif)

    network = network_from_key_value("wif", payload[0:1])
    ec = NETWORKS[network]["curve"]

    if len(payload) == ec.nsize + 2:  # compressed WIF
        compr = True
        if payload[-1] != 0x01:  # must have a trailing 0x01
            raise ValueError("Not a compressed WIF: missing trailing 0x01")
        prvkey = payload[1:-1]
    elif len(payload) == ec.nsize + 1:  # uncompressed WIF
        compr = False
        prvkey = payload[1:]
    else:
        raise ValueError(f"Wrong WIF size ({len(payload)})")

    if compressed is not None and compr != compressed:
        raise ValueError("Compression requirement mismatch")

    q = int.from_bytes(prvkey, byteorder="big")
    if not 0 < q < ec.n:
        raise ValueError(f"Private key {hex(q)} not in [1, n-1]")

    return q, network, compr


def _prvkeyinfo_from_xprv(
    xprv: BIP32Key, network: Optional[str] = None, compressed: Optional[bool] = None
) -> PrvKeyInfo:
    """Return prvkey tuple (int, compressed, network) from BIP32 xprv.

    BIP32Key is always compressed and includes network information:
    here the 'network, compressed' input parameters are passed
    only to allow consistency checks.
    """

    compressed = True if compressed is None else compressed
    if not compressed:
        raise ValueError("Uncompressed SEC / compressed BIP32 mismatch")

    if not isinstance(xprv, dict):
        xprv = bip32.deserialize(xprv)
    if xprv["key"][0] != 0:
        m = f"Not a private key: {bip32.serialize(xprv).decode()}"
        raise ValueError(m)

    q = int.from_bytes(xprv["key"][1:], byteorder="big")

    if network is not None:
        allowed_versions = xprvversions_from_network(network)
        if xprv["version"] not in allowed_versions:
            m = f"Not a key for ({network}): "
            m += f"{bip32.serialize(xprv).decode()}"
            raise ValueError(m)
        return q, network, True
    else:
        return q, network_from_xkeyversion(xprv["version"]), True


def _prvkeyinfo_from_xprvwif(
    xprvwif: BIP32Key, network: Optional[str] = None, compressed: Optional[bool] = None
) -> PrvKeyInfo:
    """Return prvkey tuple (int, compressed, network) from WIF/BIP32.

    Support WIF or BIP32 xprv.
    """

    if not isinstance(xprvwif, dict):
        try:
            return _prvkeyinfo_from_wif(xprvwif, network, compressed)
        except Exception:
            pass

    return _prvkeyinfo_from_xprv(xprvwif, network, compressed)


def prvkeyinfo_from_prvkey(
    prvkey: PrvKey, network: Optional[str] = None, compressed: Optional[bool] = None
) -> PrvKeyInfo:

    compr = True if compressed is None else compressed
    net = "mainnet" if network is None else network
    ec = NETWORKS[net]["curve"]

    if isinstance(prvkey, int):
        q = prvkey
    elif isinstance(prvkey, dict):
        return _prvkeyinfo_from_xprv(prvkey, network, compressed)
    else:
        try:
            return _prvkeyinfo_from_xprvwif(prvkey, network, compressed)
        except Exception:
            pass

        # it must octets
        prvkey = bytes_from_octets(prvkey, ec.nsize)
        q = int.from_bytes(prvkey, "big")

    if not 0 < q < ec.n:
        raise ValueError(f"Private key {hex(q).upper()} not in [1, n-1]")

    return q, net, compr



#!/usr/bin/env python3

# Copyright (C) 2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest

from btclib.curve import CurveGroup, _mult_aff
from btclib.curvegroupf import find_all_points, find_subgroup_points


class TestECFunctions(unittest.TestCase):
    def test_ecf(self):
        ec = CurveGroup(9739, 497, 1768)

        # challenge = 'Point Negation'
        P = (8045, 6936)
        S = ec.negate(P)
        S_exp = (8045, 2803)
        self.assertEqual(S, S_exp)

        # challenge = 'Point Addition'
        X = (5274, 2841)
        Y = (8669, 740)
        assert ec.add(X, Y) == (1024, 4440)
        assert ec.add(X, X) == (7284, 2107)
        P = (493, 5564)
        Q = (1539, 4742)
        R = (4403, 5202)
        S = ec.add(ec.add(ec.add(P, P), Q), R)
        ec.require_on_curve(S)
        S_exp = (4215, 2162)
        self.assertEqual(S, S_exp)

        # challenge = 'Scalar Multiplication'
        X = (5323, 5438)
        assert _mult_aff(1337, X, ec) == (1089, 6931)
        P = (2339, 2213)
        S = _mult_aff(7863, P, ec)
        ec.require_on_curve(S)
        S_exp = (9467, 2742)
        self.assertEqual(S, S_exp)

        # challenge = 'Curves and Logs'
        all_points = find_all_points(ec)
        self.assertEqual(len(all_points), 9735)
        G = (1804, 5368)
        points = find_subgroup_points(ec, G)
        self.assertEqual(len(points), 9735)
        # QA = (815, 3190)
        # nB = 1829
        # S = _mult_aff(nB, QA, ec)
        # b = S[0].to_bytes(ec.psize, 'big')
        # s = sha1(b).hexdigest()
        # print(f"{challenge}: {s}")

    def test_ecf_exceptions(self):
        ec = CurveGroup(10007, 497, 1768)

        # p (10007) is too big to count all group points
        self.assertRaises(ValueError, find_all_points, ec)
        # find_all_points(ec)

        # p (10007) is too big to count all subgroup points
        G = (2, 3265)
        self.assertRaises(ValueError, find_subgroup_points, ec, G)
        # find_subgroup_points(ec, G)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Elliptic curve point multiplication functions."""

import heapq
from typing import List, Sequence

from .alias import INFJ, JacPoint, Point
from .curve import Curve, CurveGroup, _jac_from_aff, _mult_jac
from .curves import secp256k1


def mult(m: int, Q: Point = None, ec: Curve = secp256k1) -> Point:
    """Point multiplication, implemented using 'double and add'.

    Computations use Jacobian coordinates and binary decomposition of m.
    """
    if Q is None:
        QJ = ec.GJ
    else:
        ec.require_on_curve(Q)
        QJ = _jac_from_aff(Q)

    m %= ec.n
    R = _mult_jac(m, QJ, ec)
    return ec._aff_from_jac(R)


def double_mult(u: int, H: Point, v: int, Q: Point, ec: Curve = secp256k1) -> Point:
    """Shamir trick for efficient computation of u*H + v*Q"""

    ec.require_on_curve(H)
    HJ = _jac_from_aff(H)

    ec.require_on_curve(Q)
    QJ = _jac_from_aff(Q)

    u %= ec.n
    v %= ec.n
    R = _double_mult(u, HJ, v, QJ, ec)
    return ec._aff_from_jac(R)


def _double_mult(
    u: int, HJ: JacPoint, v: int, QJ: JacPoint, ec: CurveGroup
) -> JacPoint:

    if u == 0 or HJ[2] == 0:
        return _mult_jac(v, QJ, ec)

    if v == 0 or QJ[2] == 0:
        return _mult_jac(u, HJ, ec)

    R = INFJ  # initialize as infinity point
    msb = max(u.bit_length(), v.bit_length())
    while msb > 0:
        if u >> (msb - 1):  # checking msb
            R = ec._add_jac(R, HJ)
            u -= pow(2, u.bit_length() - 1)
        if v >> (msb - 1):  # checking msb
            R = ec._add_jac(R, QJ)
            v -= pow(2, v.bit_length() - 1)
        if msb > 1:
            R = ec._add_jac(R, R)
        msb -= 1

    return R


def multi_mult(
    scalars: Sequence[int], Points: Sequence[Point], ec: CurveGroup = secp256k1
) -> Point:
    """Return the multi scalar multiplication u1*Q1 + ... + un*Qn.

    Use Bos-Coster's algorithm for efficient computation;
    the input points must be on the curve.
    """

    if len(scalars) != len(Points):
        errMsg = f"mismatch between scalar length ({len(scalars)}) and "
        errMsg += f"Points length ({len(Points)})"
        raise ValueError(errMsg)

    JPoints: List[JacPoint] = list()
    for P in Points:
        ec.require_on_curve(P)
        JPoints.append(_jac_from_aff(P))

    R = _multi_mult(scalars, JPoints, ec)

    return ec._aff_from_jac(R)


def _multi_mult(
    scalars: Sequence[int], JPoints: Sequence[JacPoint], ec: CurveGroup
) -> JacPoint:
    # source: https://cr.yp.to/badbatch/boscoster2.py

    x = list(zip([-n for n in scalars], JPoints))
    heapq.heapify(x)
    while len(x) > 1:
        np1 = heapq.heappop(x)
        np2 = heapq.heappop(x)
        n1, p1 = -np1[0], np1[1]
        n2, p2 = -np2[0], np2[1]
        p2 = ec._add_jac(p1, p2)
        n1 -= n2
        if n1 > 0:
            heapq.heappush(x, (-n1, p1))
        heapq.heappush(x, (-n2, p2))
    np1 = heapq.heappop(x)
    n1, p1 = -np1[0], np1[1]
    # n1 %= ec.n
    return _mult_jac(n1, p1, ec)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Diffie-Hellman elliptic curve key agreement scheme.

Implementation of the Diffie-Hellman key agreement scheme using
elliptic curve cryptography. A key agreement scheme is used
by two entities to establish shared keying data, which will be
later utilized e.g. in symmetric cryptographic scheme.

The two entities must agree on the elliptic curve and key derivation
function to use.
"""

from hashlib import sha256
from typing import Any, Callable

from .alias import HashF, Point
from .curve import Curve
from .curvemult import mult
from .curves import secp256k1

KDF = Callable[[bytes, int, Curve, HashF], Any]


def ansi_x963_kdf(
    z: bytes, size: int, ec: Curve = secp256k1, hf: HashF = sha256
) -> bytes:
    """Return keying data according to ANS-X9.63-KDF.

    Return a keying data octet sequence of the requested size according
    to ANS-X9.63-KDF specifications for the key derivation function.

    http://www.secg.org/sec1-v2.pdf, section 3.6.1
    """
    hsize = hf().digest_size
    assert size < hsize * (2 ** 32 - 1), "invalid"
    counter = 1
    counter_bytes = counter.to_bytes(4, byteorder="big")
    K_temp = []
    for i in range((size + 1) // hsize):
        h = hf()
        h.update(z + counter_bytes)
        K_temp.append(h.digest())
        counter += 1
        counter_bytes = counter.to_bytes(4, byteorder="big")
        i += 1
    K_bytes = b"".join(K_temp[i] for i in range(size // hsize))
    K = int.from_bytes(K_bytes, byteorder="big") >> (size - hsize)
    return K.to_bytes(ec.psize, "big")


def diffie_hellman(
    kdf: KDF, dU: int, QV: Point, size: int, ec: Curve = secp256k1, hf: HashF = sha256
) -> bytes:
    """Diffie-Hellman elliptic curve key agreement scheme.

    http://www.secg.org/sec1-v2.pdf, section 6.1
    """

    P = mult(dU, QV, ec)
    assert P[1] != 0, "invalid (INF) key"
    shared_secret = P[0]  # shared secret field element
    z = shared_secret.to_bytes(ec.psize, "big")
    return kdf(z, size, ec, hf)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import math
import secrets
import unittest

from btclib.entropy import (
    _entropy_from_indexes,
    _indexes_from_entropy,
    binstr_from_entropy,
    randbinstr,
)


class TestEntropy(unittest.TestCase):
    def test_indexes(self):
        entropy = "0"
        indexes = _indexes_from_entropy(entropy, 2048)
        self.assertEqual(indexes, [0])
        entropy = "00000000000"
        indexes = _indexes_from_entropy(entropy, 2048)
        self.assertEqual(indexes, [0])
        entropy = "000000000000"
        indexes = _indexes_from_entropy(entropy, 2048)
        self.assertEqual(indexes, [0, 0])

        test_indexes = [1268, 535, 810, 685, 433, 811, 1385, 1790, 421, 570, 567, 1313]

        entropy = _entropy_from_indexes(test_indexes, 2048)
        indexes = _indexes_from_entropy(entropy, 2048)
        self.assertEqual(indexes, test_indexes)

        test_indexes = [0, 0, 2047, 2047, 2047, 2047, 2047, 2047, 2047, 2047, 2047, 0]
        entropy = _entropy_from_indexes(test_indexes, 2048)
        indexes = _indexes_from_entropy(entropy, 2048)
        self.assertEqual(indexes, test_indexes)

        test_indexes = [0, 0, 2047, 2047, 2047, 2047, 2047, 2047, 2047, 2047, 2047, 0]
        entropy = _entropy_from_indexes(test_indexes, 2048)
        indexes = _indexes_from_entropy(entropy, 2048)
        self.assertEqual(indexes, test_indexes)

    def test_conversions(self):
        binstr_entropy = "10101011" * 32
        entropy = binstr_from_entropy(binstr_entropy)
        self.assertEqual(entropy, binstr_entropy)

        int_entropy = int(binstr_entropy, 2)
        entropy = binstr_from_entropy(int_entropy)
        self.assertEqual(entropy, binstr_entropy)

        entropy = binstr_from_entropy(bin(int_entropy))
        self.assertEqual(entropy, binstr_entropy)

        bytes_entropy = int_entropy.to_bytes(32, byteorder="big")
        entropy = binstr_from_entropy(bytes_entropy)
        self.assertEqual(entropy, binstr_entropy)

        binstr_entropy = "00101011" * 32
        entropy = binstr_from_entropy(binstr_entropy)
        self.assertEqual(entropy, binstr_entropy)

        int_entropy = int(binstr_entropy, 2)
        entropy = binstr_from_entropy(int_entropy)
        self.assertEqual(entropy, binstr_entropy)

        bytes_entropy = int_entropy.to_bytes(32, byteorder="big")
        entropy = binstr_from_entropy(bytes_entropy)
        self.assertEqual(entropy, binstr_entropy)

        binstr_entropy = "00000000" + "10101011" * 31
        entropy = binstr_from_entropy(binstr_entropy)
        self.assertEqual(entropy, binstr_entropy)

        int_entropy = int(binstr_entropy, 2)
        entropy = binstr_from_entropy(int_entropy)
        self.assertEqual(entropy, binstr_entropy)

        bytes_entropy = int_entropy.to_bytes(32, byteorder="big")
        entropy = binstr_from_entropy(bytes_entropy)
        self.assertEqual(entropy, binstr_entropy)

        # the 32 bytes integer has its leftmost bit set to 0
        int_entropy = secrets.randbits(255)
        binstr_entropy = binstr_from_entropy(int_entropy)
        self.assertEqual(len(binstr_entropy), 256)

        # 257 bits
        int_entropy = 1 << 256
        binstr_entropy = binstr_from_entropy(int_entropy)
        self.assertEqual(len(binstr_entropy), 256)

        exp_int_entropy = int_entropy >> 1
        int_entropy = int(binstr_entropy, 2)
        self.assertEqual(int_entropy, exp_int_entropy)

    def test_exceptions(self):
        binstr_entropy1 = "00011010" * 27  # 216 bits
        binstr_entropy = binstr_entropy1[2:]  # 214 bits

        entropy = binstr_from_entropy(binstr_entropy, 214)
        self.assertEqual(entropy, binstr_entropy)
        self.assertRaises(ValueError, binstr_from_entropy, binstr_entropy)
        # binstr_from_entropy(binstr_entropy)
        self.assertRaises(ValueError, binstr_from_entropy, binstr_entropy, 216)
        # binstr_from_entropy(binstr_entropy, 216)

        int_entropy = int(binstr_entropy, 2)  # 213 bits
        entropy = binstr_from_entropy(int_entropy, 214)
        self.assertEqual(entropy, binstr_entropy)
        entropy = binstr_from_entropy(int_entropy, 256)
        self.assertEqual(len(entropy), 256)
        self.assertEqual(int(entropy, 2), int_entropy)
        entropy = binstr_from_entropy(int_entropy)
        self.assertEqual(len(entropy), 224)
        self.assertEqual(int(entropy, 2), int_entropy)
        self.assertRaises(ValueError, binstr_from_entropy, -1 * int_entropy)
        # binstr_from_entropy(-1*int_entropy)

        bytes_entropy = int_entropy.to_bytes(27, byteorder="big")
        self.assertRaises(ValueError, binstr_from_entropy, bytes_entropy, 214)
        # binstr_from_entropy(bytes_entropy, 214)
        entropy = binstr_from_entropy(bytes_entropy, 216)
        self.assertEqual(entropy, binstr_entropy1)
        self.assertRaises(ValueError, binstr_from_entropy, bytes_entropy, 224)
        # binstr_from_entropy(bytes_entropy, 224)

        invalid_entropy = tuple()
        self.assertRaises(TypeError, binstr_from_entropy, invalid_entropy)

    def test_randbinstr(self):
        bits = 256
        dice_base = 20
        bits_per_roll = math.floor(math.log2(dice_base))
        base = 2 ** bits_per_roll
        roll_number = math.ceil(bits / bits_per_roll)

        rolls = [base for _ in range(roll_number)]
        binstr = randbinstr(bits, dice_base, rolls, False, False, False)
        self.assertEqual(binstr, "1" * 256)

        rolls = [base for _ in range(2 * roll_number)]
        binstr = randbinstr(bits, dice_base, rolls, False, False, False)
        self.assertEqual(binstr, "1" * 256)

        rolls = [1 for _ in range(roll_number)]
        binstr = randbinstr(bits, dice_base, rolls, False, False, False)
        self.assertEqual(binstr, "0" * 256)

        rolls = [1 for _ in range(2 * roll_number)]
        binstr = randbinstr(bits, dice_base, rolls, False, False, False)
        self.assertEqual(binstr, "0" * 256)

        rolls = [secrets.randbelow(base) + 1 for _ in range(roll_number)]
        binstr = randbinstr(bits, dice_base, rolls)
        self.assertEqual(len(binstr), 256)
        rolls = [secrets.randbelow(base) + 1 for _ in range(roll_number)]
        binstr2 = randbinstr(bits, dice_base, rolls)
        self.assertEqual(len(binstr2), 256)
        self.assertNotEqual(binstr, binstr2)

        binstr = randbinstr(bits)
        self.assertEqual(len(binstr), 256)
        binstr2 = randbinstr(bits)
        self.assertEqual(len(binstr2), 256)
        self.assertNotEqual(binstr, binstr2)

        # goes through bit lenght reduction before hashing
        rolls = [base for _ in range(roll_number + 1)]
        binstr = randbinstr(bits, dice_base, rolls)

        # Number of bits (255) must be in (128, 160, 192, 224, 256)
        self.assertRaises(ValueError, randbinstr, bits - 1, dice_base, rolls)
        # randbinstr(bits-1, dice_base, rolls)

        # too few usable [1-16] rolls, missing 2
        rolls = [secrets.randbelow(base) + 1 for _ in range(roll_number - 2)]
        self.assertRaises(ValueError, randbinstr, bits, dice_base, rolls)
        # randbinstr(bits, dice_base, rolls)

        # too few usable [1-16] rolls, missing 1
        rolls = [secrets.randbelow(base) + 1 for _ in range(roll_number)]
        rolls[1] = base + 1
        self.assertRaises(ValueError, randbinstr, bits, dice_base, rolls)
        # randbinstr(bits, dice_base, rolls)

        # invalid (21) roll, not in [1-20]
        rolls = [secrets.randbelow(base) + 1 for _ in range(roll_number)]
        rolls[1] = dice_base + 1
        self.assertRaises(ValueError, randbinstr, bits, dice_base, rolls)
        # randbinstr(bits, dice_base, rolls)

        # Invalid dice base (1): must be >= 2
        rolls = [secrets.randbelow(base) + 1 for _ in range(roll_number)]
        self.assertRaises(ValueError, randbinstr, bits, 1, rolls)
        # randbinstr(bits, 1, rolls)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

from typing import Optional, Tuple

from . import bip32
from .alias import BIP32Key, Key, Point, PrvKey, PubKey
from .curve import Curve
from .curvemult import mult
from .curves import secp256k1
from .network import (
    NETWORKS,
    curve_from_xpubversion,
    network_from_xkeyversion,
    xpubversions_from_network,
)
from .secpoint import bytes_from_point, point_from_octets
from .to_prvkey import prvkeyinfo_from_prvkey
from .utils import bytes_from_octets, hash160


def _point_from_xpub(xpub: BIP32Key, ec: Curve) -> Point:
    "Return an elliptic curve point tuple from a xpub key."

    if not isinstance(xpub, dict):
        xpub = bip32.deserialize(xpub)
    if xpub["key"][0] in (2, 3):
        ec2 = curve_from_xpubversion(xpub["version"])
        if ec != ec2:
            raise ValueError(f"ec/xpub version ({xpub['version'].hex()}) mismatch")
        return point_from_octets(xpub["key"], ec)
    raise ValueError(f"Not a public key: {xpub['key'].hex()}")


def point_from_key(key: Key, ec: Curve = secp256k1) -> Point:
    """Return a point tuple from any possible key representation.

    It supports:

    - BIP32 extended keys (bytes, string, or BIP32KeyDict)
    - SEC Octets (bytes or hex-string, with 02, 03, or 04 prefix)
    - native tuple
    """

    if isinstance(key, tuple):
        return point_from_pubkey(key, ec)
    elif isinstance(key, int):
        q, _, _ = prvkeyinfo_from_prvkey(key)
        return mult(q, ec.G, ec)
    else:
        try:
            q, net, _ = prvkeyinfo_from_prvkey(key)
        except Exception:
            pass
        else:
            if ec != NETWORKS[net]["curve"]:
                raise ValueError("Curve mismatch")
            return mult(q, ec.G, ec)

    return point_from_pubkey(key, ec)


def point_from_pubkey(P: PubKey, ec: Curve = secp256k1) -> Point:
    "Return an elliptic curve point tuple from a public key."

    if isinstance(P, tuple):
        if ec.is_on_curve(P) and P[1] != 0:
            return P
        raise ValueError(f"Not a public key: {P}")
    elif isinstance(P, dict):
        return _point_from_xpub(P, ec)
    else:
        try:
            return _point_from_xpub(P, ec)
        except Exception:
            pass

    return point_from_octets(P, ec)


# not used so far, probably useless
# def point_from_prvkey(prvkey: PrvKey, network: Optional[str] = None)->Point:
#    "Return an elliptic curve point tuple from a private key."
#
#    q, net, compr = prvkeyinfo_from_prvkey(prvkey, network)
#    ec = NETWORKS[net]['curve']
#    return mult(q, ec.G, ec)


PubKeyInfo = Tuple[bytes, str]


def _pubkeyinfo_from_xpub(
    xpub: BIP32Key, network: Optional[str] = None, compressed: Optional[bool] = None
) -> PubKeyInfo:
    """Return the pubkey tuple (SEC-bytes, network) from a BIP32 xpub.

    BIP32Key is always compressed and includes network information:
    here the 'network, compressed' input parameters are passed
    only to allow consistency checks.
    """

    compressed = True if compressed is None else compressed
    if not compressed:
        raise ValueError("Uncompressed SEC / compressed BIP32 mismatch")

    if not isinstance(xpub, dict):
        xpub = bip32.deserialize(xpub)
    if xpub["key"][0] not in (2, 3):
        m = f"Not a public key: {bip32.serialize(xpub).decode()}"
        raise ValueError(m)

    if network is not None:
        allowed_versions = xpubversions_from_network(network)
        if xpub["version"] not in allowed_versions:
            m = f"Not a key for ({network}): "
            m += f"{bip32.serialize(xpub).decode()}"
            raise ValueError(m)
        return xpub["key"], network
    else:
        return xpub["key"], network_from_xkeyversion(xpub["version"])


def pubkeyinfo_from_key(
    key: Key, network: Optional[str] = None, compressed: Optional[bool] = None
) -> PubKeyInfo:
    "Return the pub key tuple (SEC-bytes, network) from a pub/prv key."

    if isinstance(key, tuple):
        return pubkeyinfo_from_pubkey(key, network, compressed)
    elif isinstance(key, int):
        P, network = pubkeyinfo_from_prvkey(key, network, compressed)
    else:
        try:
            P, network = pubkeyinfo_from_prvkey(key, network, compressed)
        # FIXME: catch the NotPrvKeyError only
        except Exception:
            return pubkeyinfo_from_pubkey(key, network, compressed)

    return pubkeyinfo_from_pubkey(P, network, compressed)


def pubkeyinfo_from_pubkey(
    P: PubKey, network: Optional[str] = None, compressed: Optional[bool] = None
) -> PubKeyInfo:
    "Return the pub key tuple (SEC-bytes, network) from a public key."

    compr = True if compressed is None else compressed
    net = "mainnet" if network is None else network
    ec = NETWORKS[net]["curve"]

    if isinstance(P, tuple):
        return bytes_from_point(P, ec, compr), net
    elif isinstance(P, dict):
        return _pubkeyinfo_from_xpub(P, network, compressed)
    else:
        try:
            return _pubkeyinfo_from_xpub(P, network, compressed)
        except Exception:
            pass

    # it must octets
    if compressed is None:
        pubkey = bytes_from_octets(P, (ec.psize + 1, 2 * ec.psize + 1))
        compr = False
        if len(pubkey) == ec.psize + 1:
            compr = True
    else:
        size = ec.psize + 1 if compressed else 2 * ec.psize + 1
        pubkey = bytes_from_octets(P, size)
        compr = compressed

    # verify that it is a valid point
    Q = point_from_octets(pubkey, ec)

    return bytes_from_point(Q, ec, compr), net


def pubkeyinfo_from_prvkey(
    prvkey: PrvKey, network: Optional[str] = None, compressed: Optional[bool] = None
) -> PubKeyInfo:
    "Return the pub key tuple (SEC-bytes, network) from a private key."

    q, net, compr = prvkeyinfo_from_prvkey(prvkey, network, compressed)
    ec = NETWORKS[net]["curve"]
    Pub = mult(q, ec.G, ec)
    pubkey = bytes_from_point(Pub, ec, compr)
    return pubkey, net


def fingerprint(key: Key, network: Optional[str] = None) -> bytes:
    """Return the public key fingerprint from a private/public key.

    The fingerprint is the last four bytes
    of the compressed public key HASH160.
    """

    pubkey, _ = pubkeyinfo_from_key(key, network, compressed=True)
    return hash160(pubkey)[:4]



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest

from btclib.base58 import b58encode
from btclib.base58wif import wif_from_prvkey
from btclib.curves import secp256k1 as ec
from btclib.to_prvkey import prvkeyinfo_from_prvkey


class TestWif(unittest.TestCase):
    def test_wif_from_prvkey(self):
        prvkey = "0C28FCA386C7A227600B2FE50B7CAE11EC86D3BF1FBE471BE89827E19D72AA1D"
        test_vectors = [
            ["KwdMAjGmerYanjeui5SHS7JkmpZvVipYvB2LJGU1ZxJwYvP98617", "mainnet", True],
            ["cMzLdeGd5vEqxB8B6VFQoRopQ3sLAAvEzDAoQgvX54xwofSWj1fx", "testnet", True],
            ["5HueCGU8rMjxEXxiPuD5BDku4MkFqeZyd4dZ1jvhTVqvbTLvyTJ", "mainnet", False],
            ["91gGn1HgSap6CbU12F6z3pJri26xzp7Ay1VW6NHCoEayNXwRpu2", "testnet", False],
            [" KwdMAjGmerYanjeui5SHS7JkmpZvVipYvB2LJGU1ZxJwYvP98617", "mainnet", True],
            ["KwdMAjGmerYanjeui5SHS7JkmpZvVipYvB2LJGU1ZxJwYvP98617 ", "mainnet", True],
        ]
        for v in test_vectors:
            wif = wif_from_prvkey(prvkey, v[1], v[2])
            self.assertEqual(v[0].strip(), wif.decode("ascii"))
            q, network, compressed = prvkeyinfo_from_prvkey(v[0])
            self.assertEqual(q, int(prvkey, 16))
            self.assertEqual(network, v[1])
            self.assertEqual(compressed, v[2])

        # not a 32-bytes private key
        badq = 33 * b"\x02"
        self.assertRaises(ValueError, wif_from_prvkey, badq, "mainnet", True)
        # wif_from_prvkey(badq, 'mainnet', True)

        # private key not in (0, n)
        badq = ec.n
        self.assertRaises(ValueError, wif_from_prvkey, badq, "mainnet", True)
        # wif_from_prvkey(badq, 'mainnet', True)

        # Not a private key WIF: missing leading 0x80
        payload = b"\x81" + badq.to_bytes(ec.nsize, "big")
        badwif = b58encode(payload)
        self.assertRaises(ValueError, prvkeyinfo_from_prvkey, badwif)
        # prvkeyinfo_from_prvkey(badwif)

        # Not a compressed WIF: missing trailing 0x01
        payload = b"\x80" + badq.to_bytes(ec.nsize, "big") + b"\x00"
        badwif = b58encode(payload)
        self.assertRaises(ValueError, prvkeyinfo_from_prvkey, badwif)
        # prvkeyinfo_from_prvkey(badwif)

        # Not a WIF: wrong size (35)
        payload = b"\x80" + badq.to_bytes(ec.nsize, "big") + b"\x01\x00"
        badwif = b58encode(payload)
        self.assertRaises(ValueError, prvkeyinfo_from_prvkey, badwif)
        # prvkeyinfo_from_prvkey(badwif)

        # Not a WIF: private key not in (0, n)
        payload = b"\x80" + badq.to_bytes(ec.nsize, "big")
        badwif = b58encode(payload)
        self.assertRaises(ValueError, prvkeyinfo_from_prvkey, badwif)
        # prvkeyinfo_from_prvkey(badwif)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest

from btclib import bip32, slip32
from btclib.base58 import b58encode
from btclib.base58address import (
    b58address_from_h160,
    b58address_from_witness,
    h160_from_b58address,
    has_segwit_prefix,
    p2pkh,
    p2sh,
    p2wpkh_p2sh,
    p2wsh_p2sh,
)
from btclib.base58wif import wif_from_prvkey
from btclib.bech32address import p2wpkh, witness_from_b32address
from btclib.hashes import hash160_from_pubkey, hash256_from_script
from btclib.script import encode
from btclib.secpoint import bytes_from_point, point_from_octets
from btclib.to_prvkey import prvkeyinfo_from_prvkey
from btclib.to_pubkey import pubkeyinfo_from_prvkey
from btclib.utils import hash160


class TestAddresses(unittest.TestCase):
    def test_b58address_from_h160(self):
        addr = b"1PMycacnJaSqwwJqjawXBErnLsZ7RkXUAs"
        prefix, payload, network, _ = h160_from_b58address(addr)
        self.assertEqual(addr, b58address_from_h160(prefix, payload, network))

        addr = b"16UwLL9Risc3QfPqBUvKofHmBQ7wMtjvM"
        prefix, payload, network, _ = h160_from_b58address(addr)
        self.assertEqual(addr, b58address_from_h160(prefix, payload, network))

        addr = b"37k7toV1Nv4DfmQbmZ8KuZDQCYK9x5KpzP"
        prefix, payload, network, _ = h160_from_b58address(addr)
        self.assertEqual(addr, b58address_from_h160(prefix, payload, network))

        # Invalid base58 address prefix b'\xbb'
        bad_prefix = b"\xbb"
        self.assertRaises(
            ValueError, b58address_from_h160, bad_prefix, payload, network
        )
        # b58address_from_h160(bad_prefix, payload, network)

    def test_p2pkh_from_wif(self):
        seed = b"00" * 32  # better be random
        rxprv = bip32.rootxprv_from_seed(seed)
        path = "m/0h/0h/12"
        xprv = bip32.derive(rxprv, path)
        wif = wif_from_prvkey(xprv)
        self.assertEqual(wif, b"KyLk7s6Z1FtgYEVp3bPckPVnXvLUWNCcVL6wNt3gaT96EmzTKZwP")
        pubkey, _ = pubkeyinfo_from_prvkey(wif)
        address = p2pkh(pubkey)
        xpub = bip32.xpub_from_xprv(xprv)
        address2 = slip32.address_from_xpub(xpub)
        self.assertEqual(address, address2)

        self.assertRaises(ValueError, wif_from_prvkey, xpub)

    def test_p2pkh_from_pubkey(self):
        # https://en.bitcoin.it/wiki/Technical_background_of_version_1_Bitcoin_addresses
        pub = "02 " "50863ad64a87ae8a2fe83c1af1a8403cb53f53e486d8511dad8a04887e5b2352"
        addr = p2pkh(pub)
        self.assertEqual(addr, b"1PMycacnJaSqwwJqjawXBErnLsZ7RkXUAs")
        _, h160, _, _ = h160_from_b58address(addr)
        self.assertEqual(h160, hash160(pub))

        uncompr_pub = bytes_from_point(point_from_octets(pub), compressed=False)
        addr = p2pkh(uncompr_pub, compressed=False)
        self.assertEqual(addr, b"16UwLL9Risc3QfPqBUvKofHmBQ7wMtjvM")
        _, h160, _, _ = h160_from_b58address(addr)
        self.assertEqual(h160, hash160(uncompr_pub))

        # trailing/leading spaces in string
        pub = "  02 " "50863ad64a87ae8a2fe83c1af1a8403cb53f53e486d8511dad8a04887e5b2352"
        addr = p2pkh(pub)
        self.assertEqual(addr, b"1PMycacnJaSqwwJqjawXBErnLsZ7RkXUAs")
        _, h160, _, _ = h160_from_b58address(addr)
        self.assertEqual(h160, hash160(pub))

        pub = (
            "02 "
            "50863ad64a87ae8a2fe83c1af1a8403cb53f53e486d8511dad8a04887e5b2352"
            "  "
        )
        addr = p2pkh(pub)
        self.assertEqual(addr, b"1PMycacnJaSqwwJqjawXBErnLsZ7RkXUAs")

    def test_p2sh(self):
        # https://medium.com/@darosior/bitcoin-raw-transactions-part-2-p2sh-94df206fee8d
        script = [
            "OP_2DUP",
            "OP_EQUAL",
            "OP_NOT",
            "OP_VERIFY",
            "OP_SHA1",
            "OP_SWAP",
            "OP_SHA1",
            "OP_EQUAL",
        ]
        script_bytes = encode(script)
        self.assertEqual(script_bytes.hex(), "6e879169a77ca787")

        network = "mainnet"
        addr = p2sh(script_bytes, network)
        self.assertEqual(addr, b"37k7toV1Nv4DfmQbmZ8KuZDQCYK9x5KpzP")

        _, redeem_script_hash, network2, is_p2sh = h160_from_b58address(addr)
        self.assertEqual(network, network2)
        self.assertTrue(is_p2sh)
        self.assertEqual(redeem_script_hash, hash160(script_bytes))

        self.assertEqual(
            redeem_script_hash.hex(), "4266fc6f2c2861d7fe229b279a79803afca7ba34"
        )
        output_script = ["OP_HASH160", redeem_script_hash.hex(), "OP_EQUAL"]
        _ = encode(output_script)

        # address with trailing/leading spaces
        _, h160, _, _ = h160_from_b58address(" 37k7toV1Nv4DfmQbmZ8KuZDQCYK9x5KpzP ")
        self.assertEqual(redeem_script_hash, h160)

    def test_exceptions(self):

        # Invalid base58 address prefix b'\xf5'
        pubkey = (
            "02 " "50863ad64a87ae8a2fe83c1af1a8403cb53f53e486d8511dad8a04887e5b2352"
        )
        payload = b"\xf5" + hash160(pubkey)
        invalid_address = b58encode(payload)
        self.assertRaises(ValueError, h160_from_b58address, invalid_address)
        # _h160_from_b58address(invalid_address)

        # Invalid SEC pubkey length: 34-bytes
        self.assertRaises(ValueError, p2pkh, pubkey + "00")
        # p2pkh(pubkey+'00')

    def test_p2w_p2sh(self):

        pubkey = (
            "03 " "a1af804ac108a8a51782198c2d034b28bf90c8803f5a53f76276fa69a4eae77f"
        )
        h160pubkey, network = hash160_from_pubkey(pubkey)
        b58addr = p2wpkh_p2sh(pubkey, network)
        b58addr2 = b58address_from_witness(h160pubkey, network)
        self.assertEqual(b58addr2, b58addr)
        script = encode(
            ["OP_DUP", "OP_HASH160", h160pubkey, "OP_EQUALVERIFY", "OP_CHECKSIG"]
        )
        h256script = hash256_from_script(script)
        b58addr = p2wsh_p2sh(script, network)
        b58addr2 = b58address_from_witness(h256script, network)
        self.assertEqual(b58addr2, b58addr)

        # Invalid witness program length (19)
        self.assertRaises(ValueError, b58address_from_witness, h256script[:-1], network)
        # b58address_from_witness(h256script[:-1], network)

    def test_address_from_wif(self):
        # uncompressed mainnet
        wif1 = "5J1geo9kcAUSM6GJJmhYRX1eZEjvos9nFyWwPstVziTVueRJYvW"
        pubkey, network = pubkeyinfo_from_prvkey(wif1)
        b58 = p2pkh(pubkey)
        self.assertEqual(b58, b"1LPM8SZ4RQDMZymUmVSiSSvrDfj1UZY9ig")
        self.assertRaises(ValueError, p2wpkh, pubkey)
        self.assertRaises(ValueError, p2wpkh_p2sh, pubkey)

        # compressed mainnet
        wif2 = "Kx621phdUCp6sgEXPSHwhDTrmHeUVrMkm6T95ycJyjyxbDXkr162"
        pubkey, network = pubkeyinfo_from_prvkey(wif2)
        b58 = p2pkh(pubkey)
        self.assertEqual(b58, b"1HJC7kFvXHepkSzdc8RX6khQKkAyntdfkB")
        b32 = p2wpkh(pubkey)
        self.assertEqual(
            h160_from_b58address(b58)[1:], witness_from_b32address(b32)[1:]
        )
        h160 = h160_from_b58address(b58)[1]
        b = p2wpkh_p2sh(pubkey)
        self.assertEqual(hash160(b"\x00\x14" + h160), h160_from_b58address(b)[1])

        self.assertEqual(
            prvkeyinfo_from_prvkey(wif1)[0], prvkeyinfo_from_prvkey(wif2)[0]
        )

        # uncompressed testnet
        wif1 = "91gGn1HgSap6CbU12F6z3pJri26xzp7Ay1VW6NHCoEayNXwRpu2"
        pubkey, network = pubkeyinfo_from_prvkey(wif1)
        b58 = p2pkh(pubkey, network, None)
        self.assertEqual(b58, b"mvgbzkCSgKbYgaeG38auUzR7otscEGi8U7")
        self.assertRaises(ValueError, p2wpkh, pubkey)
        self.assertRaises(ValueError, p2wpkh_p2sh, pubkey)

        # compressed testnet
        wif2 = "cMzLdeGd5vEqxB8B6VFQoRopQ3sLAAvEzDAoQgvX54xwofSWj1fx"
        pubkey, network = pubkeyinfo_from_prvkey(wif2)
        b58 = p2pkh(pubkey, network, None)
        self.assertEqual(b58, b"n1KSZGmQgB8iSZqv6UVhGkCGUbEdw8Lm3Q")
        b32 = p2wpkh(pubkey, network)
        self.assertEqual(
            h160_from_b58address(b58)[1:], witness_from_b32address(b32)[1:]
        )
        h160 = h160_from_b58address(b58)[1]
        b = p2wpkh_p2sh(pubkey, network)
        self.assertEqual(hash160(b"\x00\x14" + h160), h160_from_b58address(b)[1])

        self.assertEqual(
            prvkeyinfo_from_prvkey(wif1)[0], prvkeyinfo_from_prvkey(wif2)[0]
        )

        # uncompressed mainnet, trailing/leading spaces in string
        wif1 = "  5J1geo9kcAUSM6GJJmhYRX1eZEjvos9nFyWwPstVziTVueRJYvW"
        pubkey, network = pubkeyinfo_from_prvkey(wif1)
        b58 = p2pkh(pubkey)
        self.assertEqual(b58, b"1LPM8SZ4RQDMZymUmVSiSSvrDfj1UZY9ig")
        self.assertRaises(ValueError, p2wpkh, pubkey)
        self.assertRaises(ValueError, p2wpkh_p2sh, pubkey)

        # compressed mainnet, trailing/leading spaces in string
        wif2 = "Kx621phdUCp6sgEXPSHwhDTrmHeUVrMkm6T95ycJyjyxbDXkr162  "
        pubkey, network = pubkeyinfo_from_prvkey(wif2)
        b58 = p2pkh(pubkey)
        self.assertEqual(b58, b"1HJC7kFvXHepkSzdc8RX6khQKkAyntdfkB")
        b32 = p2wpkh(pubkey)
        self.assertEqual(
            h160_from_b58address(b58)[1:], witness_from_b32address(b32)[1:]
        )
        h160 = h160_from_b58address(b58)[1]
        b = p2wpkh_p2sh(pubkey)
        self.assertEqual(hash160(b"\x00\x14" + h160), h160_from_b58address(b)[1])

    def test_has_segwit_prefix(self):
        addr = b"bc1q0hy024867ednvuhy9en4dggflt5w9unw4ztl5a"
        self.assertTrue(has_segwit_prefix(addr))
        self.assertTrue(has_segwit_prefix(addr.decode()))
        addr = b"1PMycacnJaSqwwJqjawXBErnLsZ7RkXUAs"
        self.assertFalse(has_segwit_prefix(addr))
        self.assertFalse(has_segwit_prefix(addr.decode()))


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2019-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

""" Bitcoin message signing (BMS).

Bitcoin uses a P2PKH address-based scheme for message signature: such
a signature does prove the control of the private key corresponding to
the address and, consequently, of the associated bitcoins (if any).
Message signature adopts a custom compact 65-bytes (fixed size)
serialization format (i.e. not the ASN.1 DER format used for
transactions, which would results in 71-bytes average signature).

One should never sign a vague statement that could be reused
out of the context it was intended for. Always include at least:

- name (nickname, customer id, e-mail, etc.)
- date and time
- who the message is intended for (name, business name, e-mail, etc.)
- specific purpose of the message

To mitigate the risk of signing a possibly deceiving message,
for any given message a *magic* "Bitcoin Signed Message:\\n" prefix is
added, then the hash of the resulting message is signed.

This BMS scheme relies on ECDSA,
i.e. it works with private/public key pairs, not addresses:
the address is only used to identify a key pair.
At signing time, a wallet infrastructure is required to access
the private key corresponding to a given address;
alternatively, the private key must be provided explicitly.

To verify the ECDSA signature the public key is not needed
because (EC)DSA allows public key recovery:
public keys that correctly verify the signature
can be implied from the signature itself.
In the case of the Bitcoin secp256k1 curve,
two public keys are recovered
(up to four with non-zero but negligible probability);
at verification time the address must match
that public key in the recovery set
marked as the right one at signature time.

The (r, s) DSA signature is serialized as
[1-byte recovery flag][32-bytes r][32-bytes s],
in a compact 65-bytes (fixed-size) encoding.

The serialized signature is then base64-encoded to transport it
across channels that are designed to deal with textual data.
Base64-encoding uses 10 digits, 26 lowercase characters, 26 uppercase
characters, '+' (plus sign), and '/' (forward slash).
The equal sign '=' is used as encoding end marker.

The recovery flag is used
at verification time to discriminate among recovered
public keys (and among address types in the case
of scheme extension beyond P2PKH).
Explicitly, the recovery flag value is:

    key_id + (4 if compressed else 0) + 27

where:

- key_id is the index in the [0, 3] range identifying which of the
  recovered public keys is the one associated to the address;
- compressed indicates if the address is the hash of the compressed
  public key representation
- 27 identify a P2PKH address, which is the only kind of address
  supported by Bitcoin Core;
  when the recovery flag is in the [31, 34] range of compressed
  addresses, Electrum also check for P2WPKH-P2SH and P2WPKH
  (SegWit always uses compressed public keys);
  BIP137 (Trezor) uses, instead, 35 and 39 instead of 27
  for P2WPKH-P2SH and P2WPKH (respectively).

+------+-----+-----------------------------------------------------+
|  rec | key |                                                     |
| flag |  id | address type                                        |
+======+=====+=====================================================+
|  27  |  0  | P2PKH uncompressed                                  |
+------+-----+-----------------------------------------------------+
|  28  |  1  | P2PKH uncompressed                                  |
+------+-----+-----------------------------------------------------+
|  29  |  2  | P2PKH uncompressed                                  |
+------+-----+-----------------------------------------------------+
|  30  |  3  | P2PKH uncompressed                                  |
+------+-----+-----------------------------------------------------+
|  31  |  0  | P2PKH compressed (also Electrum P2WPKH-P2SH/P2WPKH) |
+------+-----+-----------------------------------------------------+
|  32  |  1  | P2PKH compressed (also Electrum P2WPKH-P2SH/P2WPKH) |
+------+-----+-----------------------------------------------------+
|  33  |  2  | P2PKH compressed (also Electrum P2WPKH-P2SH/P2WPKH) |
+------+-----+-----------------------------------------------------+
|  34  |  3  | P2PKH compressed (also Electrum P2WPKH-P2SH/P2WPKH) |
+------+-----+-----------------------------------------------------+
|  35  |  0  | BIP137 (Trezor) P2WPKH-P2SH                         |
+------+-----+-----------------------------------------------------+
|  36  |  1  | BIP137 (Trezor) P2WPKH-P2SH                         |
+------+-----+-----------------------------------------------------+
|  37  |  2  | BIP137 (Trezor) P2WPKH-P2SH                         |
+------+-----+-----------------------------------------------------+
|  38  |  3  | BIP137 (Trezor) P2WPKH-P2SH                         |
+------+-----+-----------------------------------------------------+
|  39  |  0  | BIP137 (Trezor) P2WPKH                              |
+------+-----+-----------------------------------------------------+
|  40  |  1  | BIP137 (Trezor) P2WPKH                              |
+------+-----+-----------------------------------------------------+
|  41  |  2  | BIP137 (Trezor) P2WPKH                              |
+------+-----+-----------------------------------------------------+
|  42  |  3  | BIP137 (Trezor) P2WPKH                              |
+------+-----+-----------------------------------------------------+

This implementation endorses the Electrum approach: a signature
generated with a compressed WIF (i.e. without explicit address or
with a compressed P2PKH address) is valid also for the
P2WPKH-P2SH and P2WPKH addresses derived from the same WIF.

Nonetheless, it is possible to obtain the BIP137 behaviour if
at signing time the compressed WIF is supplemented with
a P2WPKH-P2SH or P2WPKH address:
in this case the signature will be valid only for that same
address.

https://github.com/bitcoin/bitcoin/pull/524

https://github.com/bitcoin/bips/blob/master/bip-0137.mediawiki
"""

import secrets
from base64 import b64decode, b64encode
from hashlib import sha256
from typing import Optional, Tuple

from . import dsa
from .alias import BMSig, BMSigTuple, PrvKey, String
from .base58address import h160_from_b58address, p2pkh, p2wpkh_p2sh
from .base58wif import wif_from_prvkey
from .bech32address import p2wpkh, witness_from_b32address
from .curvemult import mult
from .curves import secp256k1
from .network import NETWORKS
from .secpoint import bytes_from_point
from .to_prvkey import prvkeyinfo_from_prvkey
from .utils import hash160


def _validate_sig(rf: int, r: int, s: int) -> None:

    if rf < 27 or rf > 42:
        raise ValueError(f"Invalid recovery flag: {rf}")
    dsa._validate_sig(r, s, secp256k1)


def deserialize(sig: BMSig) -> BMSigTuple:
    """Return the verified components of the provided BSM signature.

    The address-based BSM signature can be represented
    as (rf, r, s) tuple or as base64-encoding of the compact format
    [1-byte rf][32-bytes r][32-bytes s].
    """
    if isinstance(sig, tuple):
        rf, r, s = sig
    else:
        sig = b64decode(sig)
        if len(sig) != 65:
            raise ValueError(f"Wrong signature length: {len(sig)} instead of 65")
        rf = sig[0]
        r = int.from_bytes(sig[1:33], byteorder="big")
        s = int.from_bytes(sig[33:], byteorder="big")

    _validate_sig(rf, r, s)
    return rf, r, s


def serialize(rf: int, r: int, s: int) -> bytes:
    """Return the BSM address-based signature as base64-encoding.

    First off, the signature is serialized in the
    [1-byte rf][32-bytes r][32-bytes s] compact format,
    then it is base64-encoded.
    """
    _validate_sig(rf, r, s)
    sig = bytes([rf]) + r.to_bytes(32, "big") + s.to_bytes(32, "big")
    return b64encode(sig)


def gen_keys(
    prvkey: PrvKey = None,
    network: Optional[str] = None,
    compressed: Optional[bool] = None,
) -> Tuple[bytes, bytes]:
    """Return a private/public key pair.

    The private key is a WIF, the public key is a base58 P2PKH address.
    """

    if prvkey is None:
        if network is None:
            network = "mainnet"
        ec = NETWORKS[network]["curve"]
        # q in the range [1, ec.n-1]
        q = 1 + secrets.randbelow(ec.n - 1)
        wif = wif_from_prvkey(q, network, compressed)
    else:
        wif = wif_from_prvkey(prvkey, network, compressed)

    address = p2pkh(wif)

    return wif, address


def _magic_hash(msg: String) -> bytes:

    # Electrum does strip leading and trailing spaces;
    # Bitcoin Core does not
    if isinstance(msg, str):
        msg = msg.encode()

    t = b"\x18Bitcoin Signed Message:\n" + len(msg).to_bytes(1, "big") + msg
    return sha256(t).digest()


def sign(msg: String, prvkey: PrvKey, addr: Optional[String] = None) -> BMSigTuple:
    """Generate address-based compact signature for the provided message."""

    if isinstance(addr, str):
        addr = addr.strip()
        addr = addr.encode("ascii")

    # first sign the message
    magic_msg = _magic_hash(msg)
    q, network, compressed = prvkeyinfo_from_prvkey(prvkey)
    r, s = dsa.sign(magic_msg, q)

    # now calculate the key_id
    # TODO do the match in Jacobian coordinates avoiding mod_inv
    pubkeys = dsa.recover_pubkeys(magic_msg, (r, s))
    Q = mult(q)
    # key_id is in [0, 3]
    # first two bits in rf are reserved for it
    key_id = pubkeys.index(Q)
    pubkey = bytes_from_point(Q, compressed=compressed)

    # finally, calculate the recovery flag
    if addr is None or addr == p2pkh(pubkey, network, compressed):
        rf = key_id + 27
        # third bit in rf is reserved for the 'compressed' boolean
        rf += 4 if compressed else 0
    # BIP137
    elif addr == p2wpkh_p2sh(pubkey, network):
        rf = key_id + 35
    elif addr == p2wpkh(pubkey, network):
        rf = key_id + 39
    else:
        raise ValueError("Mismatch between private key and address")

    return rf, r, s


def _to_sig(sig: BMSig) -> BMSigTuple:
    if isinstance(sig, tuple):
        rf, r, s = sig
        _validate_sig(rf, r, s)
    else:
        # it is a base64 serialized signature
        rf, r, s = deserialize(sig)
    return rf, r, s


def _verify(msg: String, addr: String, sig: BMSig) -> None:
    # Private function for test/dev purposes
    # It raises Errors, while verify should always return True or False

    rf, r, s = _to_sig(sig)

    magic_msg = _magic_hash(msg)
    # first two bits in rf are reserved for key_id
    #    key_id = 00;     key_id = 01;     key_id = 10;     key_id = 11
    # 27-27 = 000000;  28-27 = 000001;  29-27 = 000010;  30-27 = 000011
    # 31-27 = 000100;  32-27 = 000101;  33-27 = 000110;  34-27 = 000111
    # 35-27 = 001000;  36-27 = 001001;  37-27 = 001010;  38-27 = 001011
    # 39-27 = 001100;  40-27 = 001101;  41-27 = 001110;  42-27 = 001111
    key_id = rf - 27 & 0b11
    c = dsa._challenge(magic_msg, secp256k1, sha256)

    Recovered = dsa._recover_pubkey(key_id, c, r, s, secp256k1)
    Q = secp256k1._aff_from_jac(Recovered)

    try:
        _, h160, _, is_script_hash = h160_from_b58address(addr)
        is_b58 = True
    except Exception:
        _, h160, _, is_script_hash = witness_from_b32address(addr)
        is_b58 = False

    # signature is valid only if the provided address is matched
    compressed = True
    if rf < 31:
        compressed = False
    pubkey = bytes_from_point(Q, compressed=compressed)
    if is_b58:
        if is_script_hash and 30 < rf and rf < 39:  # P2WPKH-P2SH
            script_pk = b"\x00\x14" + hash160(pubkey)
            assert hash160(script_pk) == h160, "Unmatched p2wpkh-p2sh address"
        elif rf < 35:  # P2PKH
            assert hash160(pubkey) == h160, "Unmatched p2pkh address"
        else:
            m = f"Invalid recovery flag ({rf}) for base58 address ({addr!r})"
            raise ValueError(m)
    else:
        if rf > 38 or (30 < rf and rf < 35):  # P2WPKH
            assert hash160(pubkey) == h160, "Unmatched p2wpkh address"
        else:
            m = f"Invalid recovery flag ({rf}) for bech32 address ({addr!r})"
            raise ValueError(m)


def verify(msg: String, addr: String, sig: BMSig) -> bool:
    """Verify address-based compact signature for the provided message."""

    # try/except wrapper for the Errors raised by _verify
    try:
        _verify(msg, addr, sig)
    except Exception:
        return False
    else:
        return True



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import secrets
from collections import defaultdict
from hashlib import sha256 as hf  # FIXME: any hf
from typing import Dict, List, Sequence, Tuple

from .alias import Point, String
from .curvemult import double_mult, mult
from .curves import secp256k1 as ec  # FIXME: any curve
from .secpoint import bytes_from_point
from .utils import int_from_bits


def _hash(m: bytes, R: bytes, i: int, j: int) -> bytes:
    temp = m + R
    temp += i.to_bytes(4, byteorder="big") + j.to_bytes(4, byteorder="big")
    return hf(temp).digest()


PubkeyRing = Dict[int, Sequence[Point]]


def _get_msg_format(msg: bytes, pubk_rings: PubkeyRing) -> bytes:
    for pubk_ring in pubk_rings.values():
        for P in pubk_ring:
            msg += bytes_from_point(P, ec)
    return hf(msg).digest()


SValues = Dict[int, List[int]]


def sign(
    msg: String,
    ks: Sequence[int],
    sign_key_idx: Sequence[int],
    sign_keys: Sequence[int],
    pubk_rings: PubkeyRing,
) -> Tuple[bytes, SValues]:
    """Borromean ring signature - signing algorithm

    https://github.com/ElementsProject/borromean-signatures-writeup
    https://github.com/Blockstream/borromean_paper/blob/master/borromean_draft_0.01_9ade1e49.pdf

    inputs:
    - msg: message to be signed (bytes)
    - sign_key_idx: list of indexes representing each signing key per ring
    - sign_keys: list containing the whole set of signing keys (one per ring)
    - pubk_rings: dictionary of sequences representing single rings of pubkeys
    """

    if isinstance(msg, str):
        msg = msg.encode()
    m = _get_msg_format(msg, pubk_rings)

    e0bytes = m
    s: SValues = defaultdict(list)
    e: SValues = defaultdict(list)
    # step 1
    for i, (pubk_ring, j_star, k) in enumerate(
        zip(pubk_rings.values(), sign_key_idx, ks)
    ):
        keys_size = len(pubk_ring)
        s[i] = [0] * keys_size
        e[i] = [0] * keys_size
        start_idx = (j_star + 1) % keys_size
        R = bytes_from_point(mult(k), ec)
        if start_idx != 0:
            for j in range(start_idx, keys_size):
                s[i][j] = secrets.randbits(256)
                e[i][j] = int_from_bits(_hash(m, R, i, j), ec.nlen) % ec.n
                assert 0 < e[i][j] < ec.n, "sign fail: how did you do that?!?"
                T = double_mult(-e[i][j], pubk_ring[j], s[i][j], ec.G)
                R = bytes_from_point(T, ec)
        e0bytes += R
    e0 = hf(e0bytes).digest()
    # step 2
    for i, (j_star, k) in enumerate(zip(sign_key_idx, ks)):
        e[i][0] = int_from_bits(_hash(m, e0, i, 0), ec.nlen) % ec.n
        assert 0 < e[i][0] < ec.n, "sign fail: how did you do that?!?"
        for j in range(1, j_star + 1):
            s[i][j - 1] = secrets.randbits(256)
            T = double_mult(-e[i][j - 1], pubk_rings[i][j - 1], s[i][j - 1], ec.G)
            R = bytes_from_point(T, ec)
            e[i][j] = int_from_bits(_hash(m, R, i, j), ec.nlen) % ec.n
            assert 0 < e[i][j] < ec.n, "sign fail: how did you do that?!?"
        s[i][j_star] = k + sign_keys[i] * e[i][j_star]
    return e0, s


def verify(msg: String, e0: bytes, s: SValues, pubk_rings: PubkeyRing) -> bool:
    """Borromean ring signature - verification algorithm

    inputs:

    - msg: message to be signed (bytes)
    - e0: pinned e-value needed to start the verification algorithm
    - s: s-values, both real (one per ring) and forged
    - pubk_rings: dictionary of sequences representing single rings of pubkeys
    """

    if isinstance(msg, str):
        msg = msg.encode()

    # this is just a try/except wrapper for the Errors
    # raised by _verify
    try:
        return _verify(msg, e0, s, pubk_rings)
    except Exception:
        return False


def _verify(msg: bytes, e0: bytes, s: SValues, pubk_rings: PubkeyRing) -> bool:

    ring_size = len(pubk_rings)
    m = _get_msg_format(msg, pubk_rings)
    e: SValues = defaultdict(list)
    e0bytes = m
    for i in range(ring_size):
        keys_size = len(pubk_rings[i])
        e[i] = [0] * keys_size
        e[i][0] = int_from_bits(_hash(m, e0, i, 0), ec.nlen) % ec.n
        assert e[i][0] != 0, "invalid sig: how did you do that?!?"
        R = b"\0x00"
        for j in range(keys_size):
            T = double_mult(-e[i][j], pubk_rings[i][j], s[i][j], ec.G)
            R = bytes_from_point(T, ec)
            if j != len(pubk_rings[i]) - 1:
                h = _hash(m, R, i, j + 1)
                e[i][j + 1] = int_from_bits(h, ec.nlen) % ec.n
                assert e[i][j + 1] != 0, "invalid sig: how did you do that?!?"
            else:
                e0bytes += R
    e0_prime = hf(e0bytes).digest()
    return e0_prime == e0



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Entropy conversions from/to binary 0/1 string, bytes-like, and int.

Input entropy can be expressed as
binary 0/1 string, bytes-like, or integer.

Output entropy is always a binary 0/1 string.
"""

import math
import secrets
from hashlib import sha256
from typing import Iterable, List, Optional, Union

from .alias import BinStr, Entropy

_bits = 128, 160, 192, 224, 256

OneOrMoreInt = Union[int, Iterable[int]]


def binstr_from_entropy(entr: Entropy, bits: OneOrMoreInt = _bits) -> BinStr:
    """Convert the input entropy to binary 0/1 string.

    Input entropy can be expressed as
    binary 0/1 string, bytes-like, or integer:

    - a string is checked as valid 0/1 binary string of the
      required bit length, with no padding or manipulation
    - a byte sequence is checked to be of the required bit length,
      with no padding or manipulation
    - an integer is padded with leading zero bits up to the nearest
      allowed bit length; if the original bit length is longer than
      the maximum length, then only the leftmost bits are retained.

    In the case of binary 0/1 string and bytes-like
    leading zeros are not considered redundant padding.

    By default, bit-size must be 128, 160, 192, 224, or 256 bits.
    """

    if isinstance(bits, int):
        bits = (bits,)  # if a single int, make it a tuple
    bits = sorted(set(bits))  # ascending unique sorting of allowed bits

    if isinstance(entr, str):
        binstr_entr = entr.strip()
        if binstr_entr[:2] == "0b":
            binstr_entr = binstr_entr[2:]
        int(binstr_entr, 2)  # check that entr is a valid binary string
        nbits = len(binstr_entr)
        # no length adjustment
    elif isinstance(entr, bytes):
        nbits = len(entr) * 8
        int_entr = int.from_bytes(entr, "big")
        binstr_entr = bin(int_entr)[2:]  # remove '0b'
        # no length adjustment
    elif isinstance(entr, int):
        if entr < 0:
            raise ValueError(f"negative entropy ({entr})")
        binstr_entr = bin(entr)[2:]  # remove '0b'
        nbits = len(binstr_entr)
        if nbits > bits[-1]:
            # only the leftmost bits are retained
            binstr_entr = binstr_entr[: bits[-1]]
            nbits = bits[-1]
        elif nbits not in bits:
            # next allowed bit length
            nbits = next(v for i, v in enumerate(bits) if v > nbits)
    else:
        m = "entropy must be binary 0/1 string, bytes-like, or int; "
        m += f"not '{type(entr).__name__}'"
        raise TypeError(m)

    if nbits not in bits:
        raise ValueError(f"{nbits} bits entropy provided; expected: {bits}")
    return binstr_entr.zfill(nbits)  # might need padding with leading zeros


def _indexes_from_entropy(entropy: BinStr, base: int) -> List[int]:
    """Return the digit indexes for the binary 0/1 string entropy.

    Return the list of integer indexes into a digit set,
    usually a language word-list,
    for a given entropy.

    Entropy must be represented as binary 0/1 string; leading zeros
    are not considered redundant padding.
    """

    bits = len(entropy)
    int_entropy = int(entropy, 2)
    indexes = []
    while int_entropy:
        int_entropy, index = divmod(int_entropy, base)
        indexes.append(index)

    # do not lose leading zeros entropy
    bits_per_digit = int(math.log(base, 2))
    nwords = math.ceil(bits / bits_per_digit)
    while len(indexes) < nwords:
        indexes.append(0)

    return list(reversed(indexes))


def _entropy_from_indexes(indexes: List[int], base: int) -> BinStr:
    """Return the entropy from a list of word-list indexes.

    Return the entropy from a list of integer indexes into
    a given language word-list.
    """

    entropy = 0
    for index in indexes:
        entropy = entropy * base + index

    binentropy = bin(entropy)[2:]  # remove '0b'

    # do not lose leading zeros entropy
    bits_per_digit = int(math.log(base, 2))
    bits = len(indexes) * bits_per_digit
    binentropy = binentropy.zfill(bits)

    return binentropy


def randbinstr(
    bits: int,
    dice_base: int = 0,
    rolls: Optional[List[int]] = None,
    shuffle: bool = True,
    hash: bool = True,
    xor: bool = True,
) -> BinStr:
    """Return CSPRNG system entropy mixed with exogenous roll-based entropy.

    If no exogenous entropy is provided, then entropy generated with the
    system cryptographically strong pseudo-random number generator (CSPRNG)
    is returned.

    Instead, if exogenous entropy is provided, then it is possibly manipulated
    and, finally, XOR-ed with the CSPRNG system entropy.

    The optional exogenous roll-based entropy must consist of integers in the
    [1-dice_base] range; anyway, only rolls having value in the [1-base] range
    are used, with base being the highest power of 2 lower than the dice_base
    (e.g. for a traditional D6 dice, only rolls having value in [1-4] are
    used; for a D20 dice, only rolls having value in [1-16] are used; etc.).

    If provided, the exogenous roll-based entropy must supply at least the
    required number of bits.
    Rolls can be shuffled,
    resulting entropy can be hashed,
    and it is finally XOR-ed with the CSPRNG system entropy.
    If not shuffled, hashed, and/or XOR-ed, then the function returns the
    rightmost required number of bits from the unaltered exogenous entropy.
    """

    if bits not in _bits:
        raise ValueError(f"Number of bits ({bits}) must be in {_bits}")

    i = 0
    # start with the exogenously provided roll-based entropy
    if rolls is not None:
        if dice_base < 2:
            raise ValueError(f"Invalid dice base ({dice_base}): must be >= 2")
        bits_per_roll = math.floor(math.log2(dice_base))
        # used base
        base = 2 ** bits_per_roll

        if shuffle:
            secrets.SystemRandom().shuffle(rolls)

        min_roll_number = math.ceil(bits / bits_per_roll)
        for r in rolls:
            # collect only usable rolls in [1-base)]
            if 0 < r and r <= base:
                i *= base
                i += r - 1
                min_roll_number -= 1
            # reject invalid rolls not in [1-dice_base)]
            elif r < 1 or r > dice_base:
                msg = f"invalid ({r}) roll, not in [1-{dice_base}]"
                raise ValueError(msg)
        if min_roll_number > 0:
            msg = f"too few usable [1-{base}] rolls, missing {min_roll_number}"
            raise ValueError(msg)

        # hash the (possibly shuffled) exogenous entropy
        if hash:
            if i.bit_length() > 256:
                i >>= i.bit_length() - 256
            h256 = sha256(i.to_bytes(32, byteorder="big")).digest()
            i = int.from_bytes(h256, byteorder="big")

    # XOR the (possibly shuffled and/or hashed) exogenous entropy
    # with CSPRNG system entropy
    if xor:
        i ^= secrets.randbits(bits)

    # convert to binary string
    binstr = bin(i)
    # remove leading '0b'
    binstr = binstr[2:]
    # do not lose leading zeros
    binstr = binstr.zfill(bits)
    # take only the (possibly XOR-ed) rightmost bits
    return binstr[-bits:]



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""ScriptPubKey functions.

"""

from typing import List, Optional, Tuple, Union

from .alias import Key, Octets, Script, String, Token
from .hashes import hash160_from_pubkey, hash160_from_script, hash256_from_script
from .script import decode, encode
from .to_pubkey import pubkeyinfo_from_key
from .utils import bytes_from_octets

# 1. Hash/WitnessProgram from pubkey/script

# hash160_from_pubkey, hash160_from_script, and hash256_from_script
# are imported from base58address


# 2. scriptPubKey from Hash/WitnessProgram and vice versa


def scriptPubKey_from_payload(
    s_type: str,
    payloads: Union[Octets, List[Octets]],
    m: int = 0,
    lexicographic_sort: bool = True,
) -> bytes:
    "Return the requested scriptPubKey for the provided payload."

    script_type = s_type.lower()

    if (script_type == "p2ms") ^ (m != 0):
        errmsg = f"Invalid m ({m}) for {script_type} script"
        raise ValueError(errmsg)

    if isinstance(payloads, list):
        if script_type == "p2ms":
            if m < 1 or m > 16:
                raise ValueError(f"Invalid m ({m}) in m-of-n multisignature")
            if lexicographic_sort:
                payloads = sorted(payloads)
            n = len(payloads)
            if n < m or n > 16:
                raise ValueError(f"Invalid n ({n}) in {m}-of-{n} multisignature")
            script: List[Token] = [m]
            for key in payloads:
                key = bytes_from_octets(key, (33, 65))
                script.append(key)
            script.append(n)
            script.append("OP_CHECKMULTISIG")
        else:
            errmsg = f"Invalid list of Octets for {script_type} script"
            raise ValueError(errmsg)
    else:
        if script_type == "nulldata":
            payload = bytes_from_octets(payloads)
            if len(payload) > 80:
                msg = f"Invalid data lenght ({len(payload)} bytes) "
                msg += "for nulldata scriptPubKey"
                raise ValueError(msg)
            script = ["OP_RETURN", payload]
        elif script_type == "p2pk":
            payload = bytes_from_octets(payloads, (33, 65))
            script = [payload, "OP_CHECKSIG"]
        elif script_type == "p2wsh":
            payload = bytes_from_octets(payloads, 32)
            script = [0, payload]
        elif script_type == "p2pkh":
            payload = bytes_from_octets(payloads, 20)
            script = ["OP_DUP", "OP_HASH160", payload, "OP_EQUALVERIFY", "OP_CHECKSIG"]
        elif script_type == "p2sh":
            payload = bytes_from_octets(payloads, 20)
            script = ["OP_HASH160", payload, "OP_EQUAL"]
        elif script_type == "p2wpkh":
            payload = bytes_from_octets(payloads, 20)
            script = [0, payload]
        else:
            raise ValueError(f"Unknown script type: {script_type}")

    return encode(script)


Payloads = Union[bytes, List[bytes]]


def payload_from_scriptPubKey(script: Script) -> Tuple[str, Payloads, int]:
    "Return (scriptPubKey type, payload, m) from the input script."

    if isinstance(script, list):
        s = encode(script)
    else:
        s = bytes_from_octets(script)

    length = len(s)
    # p2pk
    if length == s[0] + 2 and s[0] in (0x41, 0x21) and s[-1] == 0xAC:
        # p2pk [pubkey, OP_CHECKSIG]
        # 0x41{65-byte pubkey}AC or 0x21{33-byte pubkey}AC
        return "p2pk", s[1:-1], 0
    # p2ms
    elif s[-1] == 0xAE:
        # p2ms [m, pubkeys, n, OP_CHECKMULTISIG]
        script = decode(s)
        m = int(script[0])
        if m < 1 or m > 16:
            raise ValueError(f"Invalid m ({m}) in {m}-of-n multisignature")
        n = len(script) - 3
        if n < m or n > 16:
            raise ValueError(f"Invalid n ({n}) in {m}-of-{n} multisignature")
        if n != int(script[-2]):
            errmsg = f"Keys ({n}) / n ({int(script[-2])}) mismatch "
            errmsg += "in m-of-n multisignature"
            raise ValueError(errmsg)
        keys: List[bytes] = []
        for pk in script[1:-2]:
            if isinstance(pk, int):
                raise ValueError(f"Invalid key in p2ms")
            key = bytes_from_octets(pk, (33, 65))
            keys.append(key)
        return "p2ms", keys, m
    # nulldata
    elif length <= 83 and s[0] == 0x6A:
        # nulldata [OP_RETURN, data]
        zeroone = int(length > 77)
        if s[1 + zeroone] != length - 2 - zeroone:
            raise ValueError(
                f"Wrong data lenght ({s[1+zeroone]}) in "
                f"{length}-bytes nulldata script: it should "
                f"have been {length-2-zeroone}: {decode(s)}"
            )
        if length < 77:
            # OP_RETURN, data length, data up to 74 bytes max
            # 0x6A{1 byte data-length}{data (0-74 bytes)}
            return "nulldata", s[2:], 0
        if length > 77:
            # OP_RETURN, OP_PUSHDATA1, data length, data min 75 bytes up to 80
            # 0x6A4C{1-byte data-length}{data (75-80 bytes)}
            if s[1] != 0x4C:
                raise ValueError(
                    f"Missing OP_PUSHDATA1 (0x4c) in "
                    f"{length}-bytes nulldata script, "
                    f"got {hex(s[1])} instead: {decode(s)}"
                )
            return "nulldata", s[3:], 0
        raise ValueError(f"Invalid 77 bytes OP_RETURN script length")
    # pkh
    elif length == 25 and s[:3] == b"\x76\xa9\x14" and s[-2:] == b"\x88\xac":
        # p2pkh [OP_DUP, OP_HASH160, pubkey_hash, OP_EQUALVERIFY, OP_CHECKSIG]
        # 0x76A914{20-byte pubkey_hash}88AC
        return "p2pkh", s[3 : length - 2], 0
    # sh
    elif length == 23 and s[:2] == b"\xa9\x14" and s[-1] == 0x87:
        # p2sh [OP_HASH160, script_hash, OP_EQUAL]
        # 0xA914{20-byte script_hash}87
        return "p2sh", s[2 : length - 1], 0
    # wkh
    elif length == 22 and s[:2] == b"\x00\x14":
        # p2wpkh [0, pubkey_hash]
        # 0x0014{20-byte pubkey_hash}
        return "p2wpkh", s[2:], 0
    # wsh
    elif length == 34 and s[:2] == b"\x00\x20":
        # p2wsh [0, script_hash]
        # 0x0020{32-byte script_hash}
        return "p2wsh", s[2:], 0
    else:
        errmsg = f"Unknown {len(s)}-bytes script"
        errmsg += f", starts with {s[:3].hex()}"
        errmsg += f", ends with {s[-2:].hex()}"
        errmsg += f": {decode(s)}"
        raise ValueError(errmsg)


# 1.+2. = 3. scriptPubKey from pubkey/script


def p2pk(key: Key) -> bytes:
    "Return the p2pk scriptPubKey of the provided pubkey."

    payload, _ = pubkeyinfo_from_key(key)
    return scriptPubKey_from_payload("p2pk", payload)


def p2ms(
    keys: List[Key],
    m: int,
    lexicographic_sort: bool = True,
    compressed: Optional[bool] = None,
) -> bytes:
    "Return the m-of-n multi-sig scriptPubKey of the provided pubkeys."

    pk: List[Octets] = [pubkeyinfo_from_key(p, compressed=compressed)[0] for p in keys]
    return scriptPubKey_from_payload("p2ms", pk, m, lexicographic_sort)


def nulldata(data: String) -> bytes:
    "Return the nulldata scriptPubKey of the provided data."

    if isinstance(data, str):
        data = data.encode()
    return scriptPubKey_from_payload("nulldata", data)


def p2pkh(key: Key, compressed: Optional[bool] = None) -> bytes:
    "Return the p2pkh scriptPubKey of the provided pubkey."

    pubkey_h160, _ = hash160_from_pubkey(key, compressed=compressed)
    return scriptPubKey_from_payload("p2pkh", pubkey_h160)


def p2sh(script: Script) -> bytes:
    "Return the p2sh scriptPubKey of the provided script."

    script_h160 = hash160_from_script(script)
    return scriptPubKey_from_payload("p2sh", script_h160)


def p2wpkh(key: Key) -> bytes:
    "Return the p2wpkh scriptPubKey of the provided pubkey."
    compressed = True  # needed to force check on pubkey
    pubkey_h160, _ = hash160_from_pubkey(key, compressed=compressed)
    return scriptPubKey_from_payload("p2wpkh", pubkey_h160)


def p2wsh(wscript: Script) -> bytes:
    "Return the p2wsh scriptPubKey of the provided script."

    script_h256 = hash256_from_script(wscript)
    return scriptPubKey_from_payload("p2wsh", script_h256)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import secrets
import unittest
from collections import defaultdict
from typing import Dict, List

from btclib import borromean, dsa
from btclib.alias import Point


class TestBorromeanRingSignature(unittest.TestCase):
    def test_borromean(self):
        nring = 4  # FIXME randomize; minimum number of rings?
        ring_sizes = [1 + secrets.randbelow(7) for _ in range(nring)]
        sign_key_idx = [secrets.randbelow(size) for size in ring_sizes]

        pubk_rings: Dict[int, List[Point]] = defaultdict(list)
        sign_keys: List[int] = []
        for i in range(nring):
            for j in range(ring_sizes[i]):
                priv_key, pub_key = dsa.gen_keys()
                pubk_rings[i].append(pub_key)
                if j == sign_key_idx[i]:
                    sign_keys.append(priv_key)

        msg = "Borromean ring signature"
        sig = borromean.sign(
            msg, list(range(1, 5)), sign_key_idx, sign_keys, pubk_rings
        )

        borromean._verify(msg.encode(), sig[0], sig[1], pubk_rings)
        self.assertTrue(borromean.verify(msg, sig[0], sig[1], pubk_rings))
        self.assertFalse(borromean.verify(0, sig[0], sig[1], pubk_rings))


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest

from btclib.numbertheory import mod_inv, mod_sqrt

primes = [
    2,
    3,
    5,
    7,
    11,
    13,
    17,
    19,
    23,
    29,
    31,
    37,
    41,
    43,
    47,
    53,
    59,
    61,
    67,
    71,
    73,
    79,
    83,
    89,
    97,
    101,
    103,
    107,
    109,
    113,
    2 ** 160 - 2 ** 31 - 1,
    2 ** 192 - 2 ** 32 - 2 ** 12 - 2 ** 8 - 2 ** 7 - 2 ** 6 - 2 ** 3 - 1,
    2 ** 192 - 2 ** 64 - 1,
    2 ** 224 - 2 ** 32 - 2 ** 12 - 2 ** 11 - 2 ** 9 - 2 ** 7 - 2 ** 4 - 2 - 1,
    2 ** 224 - 2 ** 96 + 1,
    2 ** 256 - 2 ** 32 - 977,
    2 ** 256 - 2 ** 224 + 2 ** 192 + 2 ** 96 - 1,
    2 ** 384 - 2 ** 128 - 2 ** 96 + 2 ** 32 - 1,
    2 ** 521 - 1,
]


class TestNumberTheory(unittest.TestCase):
    def test_mod_inv_prime(self):
        for p in primes:
            # zero has no inverse
            self.assertRaises(ValueError, mod_inv, 0, p)
            for a in range(1, min(p, 500)):  # exhausted only for small p
                inv = mod_inv(a, p)
                self.assertEqual(a * inv % p, 1)
                inv = mod_inv(a + p, p)
                self.assertEqual(a * inv % p, 1)

    def test_mod_inv(self):
        max_m = 100
        for m in range(2, max_m):
            nums = list(range(m))
            for a in nums:
                mult = [a * i % m for i in nums]
                if 1 in mult:
                    inv = mod_inv(a, m)
                    self.assertEqual(a * inv % m, 1)
                    inv = mod_inv(a + m, m)
                    self.assertEqual(a * inv % m, 1)
                else:
                    self.assertRaises(ValueError, mod_inv, a, m)

    def test_mod_sqrt(self):
        for p in primes[:30]:  # exhaustable only for small p
            hasRoot = set()
            hasRoot.add(0)
            hasRoot.add(1)
            for i in range(2, p):
                hasRoot.add(i * i % p)
            for i in range(p):
                if i in hasRoot:
                    root = mod_sqrt(i, p)
                    self.assertEqual(i, (root * root) % p)
                    root = p - root
                    self.assertEqual(i, (root * root) % p)
                    root = mod_sqrt(i + p, p)
                    self.assertEqual(i, (root * root) % p)
                else:
                    self.assertRaises(ValueError, mod_sqrt, i, p)

    def test_minus_one_quadr_res(self):
        "Ensure that if p = 3 (mod 4) then p - 1 is not a quadratic residue"
        for p in primes:
            if (p % 4) == 3:
                self.assertRaises(ValueError, mod_sqrt, p - 1, p)
            else:
                assert p == 2 or p % 4 == 1, "something is badly broken"
                root = mod_sqrt(p - 1, p)
                self.assertEqual(p - 1, root * root % p)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest

from btclib import base58address, bech32address
from btclib.base58address import (
    address_from_scriptPubKey,
    b58address_from_h160,
    b58address_from_witness,
    scriptPubKey_from_address,
)
from btclib.bech32address import b32address_from_witness
from btclib.network import NETWORKS
from btclib.script import decode, encode
from btclib.scriptpubkey import (
    nulldata,
    p2ms,
    p2pk,
    p2pkh,
    p2sh,
    p2wpkh,
    p2wsh,
    payload_from_scriptPubKey,
    scriptPubKey_from_payload,
)
from btclib.utils import hash160, sha256


class TestScriptPubKey(unittest.TestCase):
    def test_p2pk(self):

        script_type = "p2pk"

        # self-consistency
        pubkey = "02" "cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
        script = encode([pubkey, "OP_CHECKSIG"])

        # straight to the scriptPubKey
        scriptPubKey = p2pk(pubkey)
        self.assertEqual(scriptPubKey.hex(), script.hex())

        # to the scriptPubKey in two steps (through payload)
        scriptPubKey = scriptPubKey_from_payload(script_type, pubkey)
        self.assertEqual(scriptPubKey.hex(), script.hex())

        # back from the scriptPubKey to the payload
        script_type2, payload2, m2 = payload_from_scriptPubKey(scriptPubKey)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(bytes.fromhex(pubkey).hex(), payload2.hex())
        script_type2, payload2, m2 = payload_from_scriptPubKey(script)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(bytes.fromhex(pubkey).hex(), payload2.hex())

        # data -> payload in this case is invertible (no hash functions)

        # No address for p2pk script
        self.assertRaises(ValueError, address_from_scriptPubKey, scriptPubKey)
        # address_from_scriptPubKey(scriptPubKey)

        # documented test case: https://learnmeabitcoin.com/guide/p2pk
        pubkey = (
            "04"
            "ae1a62fe09c5f51b13905f07f06b99a2f7159b2225f374cd378d71302fa28414"
            "e7aab37397f554a7df5f142c21c1b7303b8a0626f1baded5c72a704f7e6cd84c"
        )

        script = (
            "4104"
            "ae1a62fe09c5f51b13905f07f06b99a2f7159b2225f374cd378d71302fa28414"
            "e7aab37397f554a7df5f142c21c1b7303b8a0626f1baded5c72a704f7e6cd84c"
            "ac"
        )
        scriptPubKey = p2pk(pubkey)
        self.assertEqual(scriptPubKey.hex(), script)

        # Invalid size: 34 bytes instead of (33, 65)
        pubkey = (
            "03" "ae1a62fe09c5f51b13905f07f06b99a2f7159b2225f374cd378d71302fa28414" "14"
        )
        self.assertRaises(ValueError, p2pk, pubkey)
        # p2pk(pubkey)

    def test_p2ms(self):

        script_type = "p2ms"

        # self-consistency
        pubkey1 = (
            "04"
            "cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
            "f7d8a473e7e2e6d317b87bafe8bde97e3cf8f065dec022b51d11fcdd0d348ac4"
        )
        pubkey2 = (
            "04"
            "61cbdcc5409fb4b4d42b51d33381354d80e550078cb532a34bfa2fcfdeb7d765"
            "19aecc62770f5b0e4ef8551946d8a540911abe3e7854a26f39f58b25c15342af"
        )
        pubkeys = [bytes.fromhex(pubkey1), bytes.fromhex(pubkey2)]
        m = 1

        # straight to the scriptPubKey
        scriptPubKey = p2ms(pubkeys, m)
        n = len(pubkeys)
        script = encode([m] + sorted(pubkeys) + [n, "OP_CHECKMULTISIG"])
        self.assertEqual(scriptPubKey.hex(), script.hex())

        # to the scriptPubKey in two steps (through payload)
        scriptPubKey = scriptPubKey_from_payload(script_type, pubkeys, m)
        self.assertEqual(scriptPubKey.hex(), script.hex())

        # back from the scriptPubKey to the payload
        script_type2, payload2, m2 = payload_from_scriptPubKey(scriptPubKey)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(m, m2)
        self.assertEqual(sorted(pubkeys), payload2)
        script_type2, payload2, m2 = payload_from_scriptPubKey(script)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(m, m2)
        self.assertEqual(sorted(pubkeys), payload2)

        # data -> payload in this case is invertible (no hash functions)

        # No address for p2ms script
        self.assertRaises(ValueError, address_from_scriptPubKey, scriptPubKey)
        # address_from_scriptPubKey(scriptPubKey)

        # documented test case: https://learnmeabitcoin.com/guide/p2ms
        pubkey1 = (
            "04"
            "cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
            "f7d8a473e7e2e6d317b87bafe8bde97e3cf8f065dec022b51d11fcdd0d348ac4"
        )
        pubkey2 = (
            "04"
            "61cbdcc5409fb4b4d42b51d33381354d80e550078cb532a34bfa2fcfdeb7d765"
            "19aecc62770f5b0e4ef8551946d8a540911abe3e7854a26f39f58b25c15342af"
        )
        pubkeys = [bytes.fromhex(pubkey1), bytes.fromhex(pubkey2)]
        m = 1
        n = 2
        script = (
            "51"  # OP_1
            "41"  # canonical 65-bytes push
            "04"
            "cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
            "f7d8a473e7e2e6d317b87bafe8bde97e3cf8f065dec022b51d11fcdd0d348ac4"
            "41"  # canonical 65-bytes push
            "04"
            "61cbdcc5409fb4b4d42b51d33381354d80e550078cb532a34bfa2fcfdeb7d765"
            "19aecc62770f5b0e4ef8551946d8a540911abe3e7854a26f39f58b25c15342af"
            "52"  # OP_2
            "ae"  # OP_CHECKMULTISIG
        )
        scriptPubKey = p2ms(pubkeys, 1, lexicographic_sort=False)
        self.assertEqual(scriptPubKey.hex(), script)

        # Impossible m>n 3-of-2 multisignature
        self.assertRaises(ValueError, p2ms, pubkeys, 3)
        # p2ms(pubkeys, 3)

        # Invalid m (0) for p2ms script
        self.assertRaises(ValueError, p2ms, pubkeys, 0)
        # p2ms(pubkeys, 0)

        # Invalid size: 66 bytes instead of 65
        self.assertRaises(ValueError, p2ms, [pubkey1 + "00", pubkey2], 1)
        # p2ms([pubkey1 + "00", pubkey2], 1)

        # Invalid n (17) in 3-of-17 multisignature
        self.assertRaises(ValueError, p2ms, [pubkey1] * 17, 3)
        # p2ms([pubkey1]*17, 3)

        # Invalid key length (66) in p2ms
        badpubkeys = sorted(pubkeys)
        badpubkeys[0] = badpubkeys[0] + b"\x00"
        self.assertRaises(
            ValueError, scriptPubKey_from_payload, script_type, badpubkeys, m
        )
        # scriptPubKey_from_payload(script_type, badpubkeys, m)

        # Invalid key length (66) in p2ms
        script = encode([m] + sorted(badpubkeys) + [n, "OP_CHECKMULTISIG"])
        self.assertRaises(ValueError, payload_from_scriptPubKey, script)
        # payload_from_scriptPubKey(script)

        # Invalid key in p2ms
        script = encode([m] + [0, pubkeys[1]] + [n, "OP_CHECKMULTISIG"])
        self.assertRaises(ValueError, payload_from_scriptPubKey, script)
        # payload_from_scriptPubKey(script)

        # Invalid m (0) for p2ms script
        self.assertRaises(
            ValueError, scriptPubKey_from_payload, script_type, pubkeys, 17
        )
        # scriptPubKey_from_payload(script_type, pubkeys, 17)

    def test_p2ms_2(self):

        pubkey1 = (
            "04"
            "cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
            "f7d8a473e7e2e6d317b87bafe8bde97e3cf8f065dec022b51d11fcdd0d348ac4"
        )
        pubkey2 = (
            "04"
            "61cbdcc5409fb4b4d42b51d33381354d80e550078cb532a34bfa2fcfdeb7d765"
            "19aecc62770f5b0e4ef8551946d8a540911abe3e7854a26f39f58b25c15342af"
        )
        pubkey3 = (
            "04"
            "79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798"
            "483ADA7726A3C4655DA4FBFC0E1108A8FD17B448A68554199C47D08FFB10D4B8"
        )
        pubkeys = [pubkey1, pubkey2, pubkey3]
        m = 1
        n = len(pubkeys)
        script = [m] + pubkeys + [n, "OP_CHECKMULTISIG"]
        payload_from_scriptPubKey(script)
        scriptPubKey_from_payload("p2ms", pubkeys, m)

        # Invalid list of Octets for p2sh script
        self.assertRaises(ValueError, scriptPubKey_from_payload, "p2sh", pubkeys, 0)
        # scriptPubKey_from_payload('p2sh', pubkeys, 0)

        # Invalid number of keys (0) in m-of-n multisignature
        script = [1, 3, "OP_CHECKMULTISIG"]
        self.assertRaises(ValueError, payload_from_scriptPubKey, script)
        # payload_from_scriptPubKey(script)

        # Keys (2) / n (3) mismatch in m-of-n multisignature
        script = [1, pubkey1, pubkey2, 3, "OP_CHECKMULTISIG"]
        self.assertRaises(ValueError, payload_from_scriptPubKey, script)
        # payload_from_scriptPubKey(script)

        # Impossible 3-of-2 multisignature
        script = [3, pubkey1, pubkey2, 2, "OP_CHECKMULTISIG"]
        self.assertRaises(ValueError, payload_from_scriptPubKey, script)
        # payload_from_scriptPubKey(script)

        # Invalid m (0) in 0-of-2 multisignature
        script = [0, pubkey1, pubkey2, 2, "OP_CHECKMULTISIG"]
        self.assertRaises(ValueError, payload_from_scriptPubKey, script)
        # payload_from_scriptPubKey(script)

        # 133-th byte in 1-of-3 multisignature payload is 0x40,
        # it should have been 0x41
        script = [1, pubkey1, pubkey2, pubkey3, 3, "OP_CHECKMULTISIG"]
        bscript = encode(script)
        script = bscript[:133] + b"\x40" + bscript[134:]
        self.assertRaises(ValueError, payload_from_scriptPubKey, script)
        # payload_from_scriptPubKey(script)

    def test_p2ms_3(self):

        # mixed compressed / uncompressed public keys
        pubkey1 = (
            "04"
            "cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
            "f7d8a473e7e2e6d317b87bafe8bde97e3cf8f065dec022b51d11fcdd0d348ac4"
        )
        pubkey2 = (
            "03" "61cbdcc5409fb4b4d42b51d33381354d80e550078cb532a34bfa2fcfdeb7d765"
        )
        pubkey3 = (
            "02" "79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798"
        )
        pubkeys = [
            bytes.fromhex(pubkey1),
            bytes.fromhex(pubkey2),
            bytes.fromhex(pubkey3),
        ]
        m = 1
        n = len(pubkeys)
        script = scriptPubKey_from_payload("p2ms", pubkeys, m)
        pubkeys.sort()
        exp_script = encode([m] + pubkeys + [n, "OP_CHECKMULTISIG"])
        self.assertEqual(script.hex(), exp_script.hex())
        script_type, payload, m2 = payload_from_scriptPubKey(script)
        self.assertEqual(script_type, "p2ms")
        self.assertEqual(m, m2)
        self.assertEqual(pubkeys, payload)

    def test_nulldata(self):

        script_type = "nulldata"

        # self-consistency
        string = "time-stamped data"
        payload = string.encode()
        script = encode(["OP_RETURN", payload])

        # straight to the scriptPubKey
        scriptPubKey = nulldata(string)
        self.assertEqual(scriptPubKey.hex(), script.hex())

        # to the scriptPubKey in two steps (through payload)
        scriptPubKey = scriptPubKey_from_payload(script_type, payload)
        self.assertEqual(scriptPubKey.hex(), script.hex())

        # back from the scriptPubKey to the payload
        script_type2, payload2, m2 = payload_from_scriptPubKey(scriptPubKey)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(payload.hex(), payload2.hex())
        script_type2, payload2, m2 = payload_from_scriptPubKey(script)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(payload.hex(), payload2.hex())

        # data -> payload in this case is invertible (no hash functions)
        self.assertEqual(payload.decode(), string)

        # No address for null data script
        self.assertRaises(ValueError, address_from_scriptPubKey, scriptPubKey)
        # address_from_scriptPubKey(scriptPubKey)

        # documented test cases: https://learnmeabitcoin.com/guide/nulldata
        string = "hello world"
        payload = string.encode()
        self.assertEqual(payload.hex(), "68656c6c6f20776f726c64")
        script = bytes.fromhex("6a0b68656c6c6f20776f726c64")
        scriptPubKey = nulldata(string)
        self.assertEqual(scriptPubKey.hex(), script.hex())
        scriptPubKey = scriptPubKey_from_payload(script_type, payload)
        self.assertEqual(scriptPubKey.hex(), script.hex())
        script_type2, payload2, m2 = payload_from_scriptPubKey(scriptPubKey)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(payload.hex(), payload2.hex())

        # documented test cases: https://learnmeabitcoin.com/guide/nulldata
        string = "charley loves heidi"
        payload = string.encode()
        self.assertEqual(payload.hex(), "636861726c6579206c6f766573206865696469")
        script = bytes.fromhex("6a13636861726c6579206c6f766573206865696469")
        scriptPubKey = nulldata(string)
        self.assertEqual(scriptPubKey.hex(), script.hex())
        scriptPubKey = scriptPubKey_from_payload(script_type, payload)
        self.assertEqual(scriptPubKey.hex(), script.hex())
        script_type2, payload2, m2 = payload_from_scriptPubKey(scriptPubKey)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(payload.hex(), payload2.hex())

        # documented test cases: https://learnmeabitcoin.com/guide/nulldata
        string = "家族も友達もみんなが笑顔の毎日がほしい"
        payload = (
            "e5aeb6e6978fe38282e58f8be98194e38282e381bfe38293e381aae38"
            "18ce7ac91e9a194e381aee6af8ee697a5e3818ce381bbe38197e38184"
        )
        self.assertEqual(string.encode().hex(), payload)
        script = bytes.fromhex(
            "6a39e5aeb6e6978fe38282e58f8be98194e38282e381bfe38293e381aae3818c"
            "e7ac91e9a194e381aee6af8ee697a5e3818ce381bbe38197e38184"
        )
        scriptPubKey = nulldata(string)
        self.assertEqual(scriptPubKey.hex(), script.hex())
        scriptPubKey = scriptPubKey_from_payload(script_type, payload)
        self.assertEqual(scriptPubKey.hex(), script.hex())
        script_type2, payload2, m2 = payload_from_scriptPubKey(scriptPubKey)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(payload, payload2.hex())

    def test_nulldata2(self):

        script_type = "nulldata"

        # max length case
        byte = b"\x00"
        for length in (0, 1, 16, 17, 74, 75, 80):
            payload = byte * length
            script = encode(["OP_RETURN", payload])

            scriptPubKey = scriptPubKey_from_payload(script_type, payload)
            self.assertEqual(scriptPubKey.hex(), script.hex())

            # back from the scriptPubKey to the payload
            script_type2, payload2, m2 = payload_from_scriptPubKey(scriptPubKey)
            self.assertEqual(script_type, script_type2)
            self.assertEqual(0, m2)
            self.assertEqual(payload.hex(), payload2.hex())
            script_type2, payload2, m2 = payload_from_scriptPubKey(decode(script))
            self.assertEqual(script_type, script_type2)
            self.assertEqual(0, m2)
            self.assertEqual(payload.hex(), payload2.hex())

    def test_nulldata3(self):

        # Invalid data lenght (81 bytes) for nulldata scriptPubKey
        payload = "00" * 81
        self.assertRaises(ValueError, scriptPubKey_from_payload, "nulldata", payload)
        # scriptPubKey_from_payload('nulldata', payload)

        # Wrong data lenght (32) in 35-bytes nulldata script:
        # it should have been 33
        script = encode(["OP_RETURN", b"\x00" * 33])
        script = script[:1] + b"\x20" + script[2:]
        self.assertRaises(ValueError, payload_from_scriptPubKey, script)
        # payload_from_scriptPubKey(script)

        # Wrong data lenght (32) in 83-bytes nulldata script:
        # it should have been 80
        script = encode(["OP_RETURN", b"\x00" * 80])
        script = script[:2] + b"\x20" + script[3:]
        self.assertRaises(ValueError, payload_from_scriptPubKey, script)
        # payload_from_scriptPubKey(script)

        # Missing OP_PUSHDATA1 (0x4c) in 83-bytes nulldata script,
        # got 0x20 instead
        script = encode(["OP_RETURN", b"\x00" * 80])
        script = script[:1] + b"\x20" + script[2:]
        self.assertRaises(ValueError, payload_from_scriptPubKey, script)
        # payload_from_scriptPubKey(script)

        # Invalid 77 bytes OP_RETURN script length
        script = b"\x6A" + b"\x4B" * 76
        self.assertRaises(ValueError, payload_from_scriptPubKey, script)
        # payload_from_scriptPubKey(script)

    def test_p2pkh(self):

        script_type = "p2pkh"

        # self-consistency
        pubkey = (
            "04"
            "cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
            "f7d8a473e7e2e6d317b87bafe8bde97e3cf8f065dec022b51d11fcdd0d348ac4"
        )
        payload = hash160(pubkey)
        script = encode(
            ["OP_DUP", "OP_HASH160", payload, "OP_EQUALVERIFY", "OP_CHECKSIG"]
        )

        # straight to the scriptPubKey
        scriptPubKey = p2pkh(pubkey)
        self.assertEqual(scriptPubKey.hex(), script.hex())

        # to the scriptPubKey in two steps (through payload)
        scriptPubKey = scriptPubKey_from_payload(script_type, payload)
        self.assertEqual(scriptPubKey.hex(), script.hex())

        # back from the scriptPubKey to the payload
        script_type2, payload2, m2 = payload_from_scriptPubKey(scriptPubKey)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(payload.hex(), payload2.hex())
        script_type2, payload2, m2 = payload_from_scriptPubKey(script)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(payload.hex(), payload2.hex())

        # data -> payload is not invertible (hash functions)

        # address
        network = "mainnet"
        address = base58address.p2pkh(pubkey, network)
        address2 = address_from_scriptPubKey(scriptPubKey, network)
        self.assertEqual(address, address2)
        prefix = NETWORKS[network]["p2pkh"]
        address2 = b58address_from_h160(prefix, payload, network)
        self.assertEqual(address, address2)

        scriptPubKey2, network2 = scriptPubKey_from_address(address)
        self.assertEqual(scriptPubKey2, scriptPubKey)
        self.assertEqual(network2, network)

        # documented test case: https://learnmeabitcoin.com/guide/p2pkh
        payload = "12ab8dc588ca9d5787dde7eb29569da63c3a238c"
        script = "76a91412ab8dc588ca9d5787dde7eb29569da63c3a238c88ac"
        scriptPubKey = scriptPubKey_from_payload(script_type, payload)
        self.assertEqual(scriptPubKey.hex(), script)
        network = "mainnet"
        address = b"12higDjoCCNXSA95xZMWUdPvXNmkAduhWv"
        address2 = address_from_scriptPubKey(scriptPubKey, network)
        self.assertEqual(address, address2)
        scriptPubKey2, network2 = scriptPubKey_from_address(address)
        self.assertEqual(scriptPubKey2, scriptPubKey)
        self.assertEqual(network2, network)

        # Invalid size: 11 bytes instead of 20
        self.assertRaises(ValueError, scriptPubKey_from_payload, "00" * 11, "p2pkh")
        # p2pkh("00"*11)

    def test_p2sh(self):

        script_type = "p2sh"

        # self-consistency
        pubkey = "02" "cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
        pubkey_hash = hash160(pubkey)
        redeem_script = scriptPubKey_from_payload("p2pkh", pubkey_hash)
        payload = hash160(redeem_script)
        script = encode(["OP_HASH160", payload, "OP_EQUAL"])

        # straight to the scriptPubKey
        scriptPubKey = p2sh(redeem_script)
        self.assertEqual(scriptPubKey.hex(), script.hex())

        # to the scriptPubKey in two steps (through payload)
        scriptPubKey = scriptPubKey_from_payload(script_type, payload)
        self.assertEqual(scriptPubKey.hex(), script.hex())

        # back from the scriptPubKey to the payload
        script_type2, payload2, m2 = payload_from_scriptPubKey(scriptPubKey)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(payload.hex(), payload2.hex())
        script_type2, payload2, m2 = payload_from_scriptPubKey(script)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(payload.hex(), payload2.hex())

        # data -> payload is not invertible (hash functions)

        # address
        network = "mainnet"
        address = base58address.p2sh(decode(redeem_script), network)
        address2 = address_from_scriptPubKey(scriptPubKey, network)
        self.assertEqual(address, address2)
        prefix = NETWORKS[network]["p2sh"]
        address2 = b58address_from_h160(prefix, payload, network)
        self.assertEqual(address, address2)

        scriptPubKey2, network2 = scriptPubKey_from_address(address)
        self.assertEqual(scriptPubKey2, scriptPubKey)
        self.assertEqual(network2, network)

        # documented test case: https://learnmeabitcoin.com/guide/p2sh
        payload = "748284390f9e263a4b766a75d0633c50426eb875"
        script = "a914748284390f9e263a4b766a75d0633c50426eb87587"
        scriptPubKey = scriptPubKey_from_payload(script_type, payload)
        self.assertEqual(scriptPubKey.hex(), script)
        network = "mainnet"
        address = b"3CK4fEwbMP7heJarmU4eqA3sMbVJyEnU3V"
        address2 = address_from_scriptPubKey(scriptPubKey, network)
        self.assertEqual(address, address2)
        scriptPubKey2, network2 = scriptPubKey_from_address(address)
        self.assertEqual(scriptPubKey2, scriptPubKey)
        self.assertEqual(network2, network)

        # Invalid size: 21 bytes instead of 20
        self.assertRaises(ValueError, scriptPubKey_from_payload, "00" * 21, "p2sh")
        # scriptPubKey_from_payload("00"*21, 'p2sh')

    def test_p2ms_p2sh(self):
        "BIP67 test vectors https://en.bitcoin.it/wiki/BIP_0067"

        test_vectors = {
            0: [
                [
                    (
                        "022df8750480ad5b26950b25c7ba79d3e37d75f640f8e5d9bcd5b1"
                        "50a0f85014da"
                    ),
                    (
                        "03e3818b65bcc73a7d64064106a859cc1a5a728c4345ff0b641209"
                        "fba0d90de6e9"
                    ),
                    (
                        "021f2f6e1e50cb6a953935c3601284925decd3fd21bc4457125768"
                        "73fb8c6ebc18"
                    ),
                ],
                b"3Q4sF6tv9wsdqu2NtARzNCpQgwifm2rAba",
            ],
            1: [
                [
                    (
                        "02ff12471208c14bd580709cb2358d98975247d8765f92bc25eab3"
                        "b2763ed605f8"
                    ),
                    (
                        "02fe6f0a5a297eb38c391581c4413e084773ea23954d93f7753db7"
                        "dc0adc188b2f"
                    ),
                ],
                b"39bgKC7RFbpoCRbtD5KEdkYKtNyhpsNa3Z",
            ],
            2: [
                [
                    (
                        "02632b12f4ac5b1d1b72b2a3b508c19172de44f6f46bcee50ba33f"
                        "3f9291e47ed0"
                    ),
                    (
                        "027735a29bae7780a9755fae7a1c4374c656ac6a69ea9f3697fda6"
                        "1bb99a4f3e77"
                    ),
                    (
                        "02e2cc6bd5f45edd43bebe7cb9b675f0ce9ed3efe613b177588290"
                        "ad188d11b404"
                    ),
                ],
                b"3CKHTjBKxCARLzwABMu9yD85kvtm7WnMfH",
            ],
            3: [
                [
                    (
                        "030000000000000000000000000000000000004141414141414141"
                        "414141414141"
                    ),
                    (
                        "020000000000000000000000000000000000004141414141414141"
                        "414141414141"
                    ),
                    (
                        "020000000000000000000000000000000000004141414141414141"
                        "414141414140"
                    ),
                    (
                        "030000000000000000000000000000000000004141414141414141"
                        "414141414140"
                    ),
                ],
                b"32V85igBri9zcfBRVupVvwK18NFtS37FuD",
            ],
            4: [
                [
                    (
                        "022df8750480ad5b26950b25c7ba79d3e37d75f640f8e5d9bcd5b1"
                        "50a0f85014da"
                    ),
                    (
                        "03e3818b65bcc73a7d64064106a859cc1a5a728c4345ff0b641209"
                        "fba0d90de6e9"
                    ),
                    (
                        "021f2f6e1e50cb6a953935c3601284925decd3fd21bc4457125768"
                        "73fb8c6ebc18"
                    ),
                ],
                b"3Q4sF6tv9wsdqu2NtARzNCpQgwifm2rAba",
            ],
        }

        m = 2
        for i in test_vectors:
            keys, address = test_vectors[i]
            errmsg = f"Test vector #{int(i)}"
            script = p2ms(keys, m)
            addr = base58address.p2sh(script)
            self.assertEqual(addr, address, errmsg)

            script = scriptPubKey_from_payload("p2ms", keys, m)
            addr = base58address.p2sh(script)
            self.assertEqual(addr, address, errmsg)

            script_type, payload, m2 = payload_from_scriptPubKey(script)
            self.assertEqual(script_type, "p2ms", errmsg)
            for key, k in zip(sorted(keys), payload):
                self.assertEqual(key, k.hex(), errmsg)
            self.assertEqual(m2, m, errmsg)

    def test_p2wpkh(self):

        script_type = "p2wpkh"

        # self-consistency
        pubkey = "02" "cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
        payload = hash160(pubkey)
        script = encode([0, payload])

        # straight to the scriptPubKey
        scriptPubKey = p2wpkh(pubkey)
        self.assertEqual(scriptPubKey.hex(), script.hex())

        # to the scriptPubKey in two steps (through payload)
        scriptPubKey = scriptPubKey_from_payload(script_type, payload)
        self.assertEqual(scriptPubKey.hex(), script.hex())

        # back from the scriptPubKey to the payload
        script_type2, payload2, m2 = payload_from_scriptPubKey(scriptPubKey)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(payload.hex(), payload2.hex())
        script_type2, payload2, m2 = payload_from_scriptPubKey(script)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(payload.hex(), payload2.hex())

        # data -> payload is not invertible (hash functions)

        # bech32 address
        network = "mainnet"
        address = bech32address.p2wpkh(pubkey, network)
        address2 = address_from_scriptPubKey(scriptPubKey, network)
        self.assertEqual(address, address2)
        address2 = b32address_from_witness(0, payload, network)
        self.assertEqual(address, address2)

        scriptPubKey2, network2 = scriptPubKey_from_address(address)
        self.assertEqual(scriptPubKey2, scriptPubKey)
        self.assertEqual(network2, network)

        # p2sh-wrapped base58 address
        address = base58address.p2wpkh_p2sh(pubkey, network)
        address2 = b58address_from_witness(payload, network)
        self.assertEqual(address, address2)

    def test_p2wsh(self):

        script_type = "p2wsh"

        # self-consistency
        pubkey = "02" "cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
        pubkey_hash = hash160(pubkey)
        redeem_script = scriptPubKey_from_payload("p2pkh", pubkey_hash)
        payload = sha256(redeem_script)
        script = encode([0, payload])

        # straight to the scriptPubKey
        scriptPubKey = p2wsh(decode(redeem_script))
        self.assertEqual(scriptPubKey.hex(), script.hex())

        # to the scriptPubKey in two steps (through payload)
        scriptPubKey = scriptPubKey_from_payload(script_type, payload)
        self.assertEqual(scriptPubKey.hex(), script.hex())

        # back from the scriptPubKey to the payload
        script_type2, payload2, m2 = payload_from_scriptPubKey(scriptPubKey)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(payload.hex(), payload2.hex())
        script_type2, payload2, m2 = payload_from_scriptPubKey(script)
        self.assertEqual(script_type, script_type2)
        self.assertEqual(0, m2)
        self.assertEqual(payload.hex(), payload2.hex())

        # data -> payload is not invertible (hash functions)

        # bech32 address
        network = "mainnet"
        address = bech32address.p2wsh(redeem_script, network)
        address2 = address_from_scriptPubKey(scriptPubKey, network)
        self.assertEqual(address, address2)
        address2 = b32address_from_witness(0, payload, network)
        self.assertEqual(address, address2)

        scriptPubKey2, network2 = scriptPubKey_from_address(address)
        self.assertEqual(scriptPubKey2, scriptPubKey)
        self.assertEqual(network2, network)

        # p2sh-wrapped base58 address
        address = base58address.p2wsh_p2sh(redeem_script, network)
        address2 = b58address_from_witness(payload, network)
        self.assertEqual(address, address2)

    def test_exceptions(self):

        # Invalid size: 11 bytes instead of 20
        self.assertRaises(ValueError, scriptPubKey_from_payload, "00" * 11, "p2wpkh")
        # scriptPubKey_from_payload("00"*11, 'p2wpkh')

        # Invalid size: 33 bytes instead of 32
        self.assertRaises(ValueError, scriptPubKey_from_payload, "00" * 33, "p2wsh")
        # scriptPubKey_from_payload("00"*33, 'p2wsh')

        # Unknown script
        script = [16, 20 * b"\x00"]
        self.assertRaises(ValueError, address_from_scriptPubKey, script)
        # address_from_scriptPubKey(script)

        # Unhandled witness version (16)
        addr = b32address_from_witness(16, 20 * b"\x00")
        self.assertRaises(ValueError, scriptPubKey_from_address, addr)
        # scriptPubKey_from_address(addr)


def test_CLT():

    network = "mainnet"

    vault_pubkeys = [b"\x00" * 33, b"\x11" * 33, b"\x22" * 33]
    recovery_pubkeys = [b"\x77" * 33, b"\x88" * 33, b"\x99" * 33]
    redeem_script = encode(
        [
            "OP_IF",
            2,
            *vault_pubkeys,
            3,
            "OP_CHECKMULTISIG",
            "OP_ELSE",
            500,
            "OP_CHECKLOCKTIMEVERIFY",
            "OP_DROP",
            2,
            *recovery_pubkeys,
            3,
            "OP_CHECKMULTISIG",
            "OP_ENDIF",
        ]
    )
    payload = sha256(redeem_script)
    script = "00207b5310339c6001f75614daa5083839fa54d46165f6c56025cc54d397a85a5708"

    scriptPubKey = p2wsh(redeem_script)
    assert scriptPubKey.hex() == script
    scriptPubKey = scriptPubKey_from_payload("p2wsh", payload)
    assert scriptPubKey.hex() == script

    address = (
        "bc1q0df3qvuuvqqlw4s5m2jsswpelf2dgct97mzkqfwv2nfe02z62uyq7n4zjj"
    ).encode()
    address2 = address_from_scriptPubKey(scriptPubKey, network)
    assert address == address2
    assert address == address2
    address2 = b32address_from_witness(0, payload, network)
    assert address == address2


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover
